(* File: mizartypecheckermain.mli *)
(* ported from satallaxmain.mli Aug 2017 *)

exception InputHelpError of string
exception InputError of string

val help_lines : string list
val mizartypechecker_main : bool -> string -> unit
val mizartypechecker_elaborate : string -> unit

open Treetype
open Syntax

type mizar_elab_res =
    BadParse
  | MizarTimeout
  | IncompleteArgs
  | CheckingFailure
  | ElaborationSuccess
  | DontKnow
					  
type gentp = Stp of stp | Mtp of trm list

(***
(string * trm) list is the current context under which the parse tree is being interpreted (local names and their corresponding elaboration)
For each local context there is different info (unknown names/elaboration).

First component (string list * bool) list) is an association list associating
  a local context with a bool which is true if the tree has an unknown name in it
  (relative to the "known" names in the given context). Currently we never
  try to elaborate if there are any unknown names.
Second component is an association list associating a local context with
  a simply typed term resulting from elaboration with gentp (either a stp, or a list of mizar types if simple type is $i)
  and also a (string * trm list) which probably records info about contstraints for unknown names (currently always [] since we don't work if there are unknown names).
***)
type mizrecinfo = ((string list * bool) list) * ((string * trm) list * (trm * gentp * (string * trm list) list)) list

val mizar_elab_and_test : float -> (identifier,mizrecinfo option) rich_grammar_tree_type -> mizar_elab_res * float
val mizar_elab_and_test_allsub : float -> (identifier,mizrecinfo option) rich_grammar_tree_type -> unit

val mizar_tree_to_thf : (identifier,mizrecinfo option) rich_grammar_tree_type -> string 

val load_mizar_knowledge : string -> unit

(*** no line breaks in string ***)
val string_of_rich_grammar_tree_type : (identifier,mizrecinfo option) rich_grammar_tree_type -> string
val list_of_strings_of_formal_reps_of_rich_grammar_tree_type : (identifier,mizrecinfo option) rich_grammar_tree_type -> string list
