open Syntax

let debug = false
let maxelaborateiters = ref 7
let elaboratelim = ref 7
let elaboratetimeout = ref 10.0
let acceptcommonmizartypes = ref false
let raiseunknownname = ref false (*** changed the default value to false, to allow the checker/elaborator to succeed even when there are unknown names simply by assuming the unknown names have whatever types required; if set to true the checker/elaborator will raise an unknown name exception if an unknown name is encountered leading to DontKnow ***)
let main_queried_sorts : (trm,unit) Hashtbl.t = Hashtbl.create 100
let main_assumed_sorts : (trm,unit) Hashtbl.t = Hashtbl.create 100

let assume_needed_sorts = ref 0
exception NeedToAssume

let unknown_name_sort_p p frl =
  match neg_body p with
  | Some(q) ->
      begin
	match head_spine q with
	| (_,(a::_)) ->
	    let cl = consts_of_trm [] a in
	    begin
	      try
		ignore (List.find
			  (fun (y,_) -> List.mem_assoc y frl)
			  cl);
		true
	      with Not_found -> false
	    end
	| _ -> false
      end
  | None ->
      begin
	match head_spine p with
	| (_,(a::_)) ->
	    let cl = consts_of_trm [] a in
	    begin
	      try
		ignore (List.find
			  (fun (y,_) -> List.mem_assoc y frl)
			  cl);
		true
	      with Not_found -> false
	    end
	| _ -> false
      end

let m1_hidden_p p =
  match neg_body p with
  | Some(_) -> false
  | None ->
      begin
	match head_spine p with
	| (Name(h,_),_) when h = "m1_hidden" -> true
	| _ -> false
      end

let count_assumed frl =
  let num_assumed = ref 0 in
  Hashtbl.iter (fun p _ -> if not (unknown_name_sort_p p frl) (* && not (m1_hidden_p p) *) then incr num_assumed) main_assumed_sorts;
  !num_assumed
  
let count_queried frl =
  let num_queried = ref 0 in
  Hashtbl.iter (fun p _ -> if not (unknown_name_sort_p p frl) (* && not (m1_hidden_p p) *) then incr num_queried) main_queried_sorts;
  !num_queried

exception UnknownName of string

let hounif1negflex = ref false

(*** copied from match.ml to modify for cong clos BEGIN ***)
(*** For Terms with Meta Vars and Meta Type Vars ***)
exception NotGround
exception PatternMatchFailed


let emptyd : (string,trm) Hashtbl.t = Hashtbl.create 1

type metatrm =
   | MGround of trm (*** Assume this is DB, Name, or a logical constant ***)
   | MVar of (int * metatrm option ref) * metatrm list
   | MLam of stp * metatrm
   | MAp of metatrm * metatrm

let mneg m = MAp(MAp(MGround(Imp),m),MGround(False))

let mneg_p m =
  match m with
  | MAp(MAp(MGround(Imp),_),MGround(False)) -> true
  | _ -> false

type ctx = stp list
type dpair = ctx * metatrm * trm * stp
type evar = int * metatrm option ref

let rec id_subst i gamma =
  match gamma with
  | (a::gammar) -> ((MGround(DB(i,a)))::(id_subst (i + 1) gammar))
  | [] -> []

let evarcount = ref 0

let string_of_evar (e,_) =
  "?" ^ (string_of_int e)

let rec new_evar v gamma a : evar * metatrm =
  match a with
  | Ar(a1,a2) ->
      let (x,xs) = new_evar v (a1::gamma) a2 in
      (x,MLam(a1,xs))
  | _ ->
      let x = (!evarcount,ref None) in
      incr evarcount;
      let s = id_subst 0 gamma in
      (x,MVar(x,s))

let rec copy_evar v x : evar =
  let y = (!evarcount,ref None) in
  incr evarcount;
  y

let lpar p = if p then "(" else ""

let rpar p = if p then ")" else ""

let rec metatrm_str_rec m p =
  match m with
  | MGround(m0) -> "<" ^ (trm_str m0) ^ ">"
  | MVar((x,r),s) ->
      begin
	match !r with
	| None ->
	    (lpar p) ^ "?" ^ (string_of_int x) ^ "[" ^ (String.concat "," (List.map (fun m -> metatrm_str_rec m false) s)) ^ "]" ^ (rpar p)
	| Some(m) ->
	    (lpar p) ^ "?" ^ (string_of_int x) ^ "{" ^ (metatrm_str_rec m false) ^ "}[" ^ (String.concat "," (List.map (fun m -> metatrm_str_rec m false) s)) ^ "]" ^ (rpar p)
      end
  | MLam(a,m1) -> (lpar p) ^ "\\_:" ^ (stp_str a) ^ "." ^ (metatrm_str_rec m1 false) ^ (rpar p)
  | MAp(m1,m2) ->
      begin
	match m1 with
	| MLam _ -> (lpar p) ^ (metatrm_str_rec m1 true) ^ " " ^ (metatrm_str_rec m2 true) ^ (rpar p)
	| _ -> (lpar p) ^ (metatrm_str_rec m1 false) ^ " " ^ (metatrm_str_rec m2 true) ^ (rpar p)
      end

let metatrm_str m = metatrm_str_rec m false
  
let rec to_meta m =
  match m with
  | Ap(m1,m2) -> MAp(to_meta m1,to_meta m2)
  | Lam(a,m1) -> MLam(a,to_meta m1)
  | _ -> MGround(m)

let rec meta_to_ground_rec m tau =
  match m with
  | MGround(m1) -> simulsubst m1 tau
  | MVar((e,x),sigma) ->
      begin
	match (!x) with
	| None -> raise NotGround
	| Some n -> meta_to_ground_rec n (List.map (fun p -> meta_to_ground_rec p tau) sigma)
      end
  | MLam(a,m1) -> Lam(a,meta_to_ground_rec m1 ((DB(0,a))::(List.map (fun n -> shift n 0 1) tau)))
  | MAp(m1,m2) -> Ap(meta_to_ground_rec m1 tau,meta_to_ground_rec m2 tau)
	
let meta_to_ground d m =
   norm d (meta_to_ground_rec m [])

let rec metashift m i j =
  match m with
    MGround(DB(k,_)) when k < i -> m
  | MGround(DB(k,a)) -> MGround(DB(k + j,a))
  | MLam(a1,m2) -> MLam(a1,metashift m2 (i + 1) j)
  | MAp(m1,m2) -> MAp(metashift m1 i j,metashift m2 i j)
  | _ -> m

let rec metasubst m i n =
  match m with
    MGround(DB(k,_)) when k < i -> m
  | MGround(DB(k,_)) when k = i -> n
  | MGround(DB(k,a)) -> MGround(DB(k - 1,a))
  | MGround(_) -> m
  | MLam(a1,m2) -> MLam(a1,metasubst m2 (i + 1) (metashift n 0 1))
  | MAp(m1,m2) -> MAp(metasubst m1 i n,metasubst m2 i n)
  | MVar(x,sigma) -> MVar(x,List.map (fun m0 -> metasubst m0 i n) sigma)

let gen_mlam_body a m =
  match m with
  | MLam(_,m1) -> m1
  | MVar(x,sigma) -> raise (Failure ("Metavar of function type? " ^ (string_of_evar x)))
  | _ -> MAp(metashift m 0 1,MGround(DB(0,a)))

let rec metatermNotFreeIn m i =
  match m with
  | MGround(DB(j,_)) when i = j -> false
  | MVar(x,ml) -> metatermNotFreeInL ml i
  | MAp(m1,m2) -> (metatermNotFreeIn m1 i) && (metatermNotFreeIn m2 i)
  | MLam(a,m1) -> metatermNotFreeIn m1 (i + 1)
  | _ -> true
and metatermNotFreeInL ml i =
  match ml with
  | (m::mr) -> (metatermNotFreeIn m i) && (metatermNotFreeInL mr i)
  | [] -> true

let rec meta_simulsubst_meta m tau =
   match m with
   | MGround(DB(k,_)) -> List.nth tau k
   | MGround(m1) -> m
   | MVar((e,x),sigma) ->
       begin
	 match (!x) with
	 | None ->
	     MVar((e,x),List.map (fun p -> meta_simulsubst_meta p tau) sigma)
	 | Some n ->
	     meta_simulsubst_meta n (List.map (fun p -> meta_simulsubst_meta p tau) sigma)
       end
   | MLam(a,m1) -> MLam(a,meta_simulsubst_meta m1 ((MGround(DB(0,a)))::(List.map (fun n -> metashift n 0 1) tau)))
   | MAp(m1,m2) -> MAp(meta_simulsubst_meta m1 tau,meta_simulsubst_meta m2 tau)

let rec metanorm1 m =
  match m with
  | MGround(m1) ->
      let (n1,b) = onlybetanorm1 m1 in
      (MGround(n1),b)
  | MAp(MLam(a,m1),m2) -> (* beta *)
      let (n1,_) = metanorm1 m1 in
      let (n2,_) = metanorm1 m2 in
      (metasubst n1 0 n2,true)
(***  | MLam(_,MAp(m1,MGround(DB(0,_)))) when (metatermNotFreeIn m1 0) -> (* eta *) (*** Removed eta here.  The way dpairs at functional type are handled, they will be more or less eta expanded on the fly anyway.  Hence there is no point in wasting time doing this. - Chad, Jan 10, 2011 ***)
      (metashift m1 0 (- 1),true) ***)
	(*** dneg ***)
  | MAp(MAp(MGround(Imp),MAp(MAp(MGround(Imp),m1),MGround(False))),MGround(False)) -> (* double negation reduce *)
      let (n1,_) = metanorm1 m1 in
      (n1,true)
  | MAp(m1,m2) ->
      let (n1,b1) = metanorm1 m1 in
      let (n2,b2) = metanorm1 m2 in
      if (b1 || b2) then
	(MAp(n1,n2),true)
      else
	(m,false)
  | MLam(a1,m1) ->
      let (n1,b1) = metanorm1 m1 in
      if b1 then
	(MLam(a1,n1),true)
      else
	(m,false)
  | MVar((e,x),sigma) ->
      begin
	match (!x) with
	| None ->
	    let (sigmar,sigmab) = metasubstnorm1 sigma in
	    if sigmab then
	      (MVar((e,x),sigmar),true)
	    else
	      (m,false)
	| Some m1 ->
	    (meta_simulsubst_meta m1 sigma,true)
      end
and metasubstnorm1 sigma =
  match sigma with
  | (m1::sigmar) ->
      let (n1,b1) = metanorm1 m1 in
      let (sigma2,b2) = metasubstnorm1 sigmar in
      if (b1 || b2) then
	(n1::sigma2,true)
      else
	(sigma,false)
  | [] -> ([],false)

(* beta-dneg *)
let rec metanorm m =
  let (m1,reduced) = metanorm1 m in
  if reduced then (metanorm m1) else m

let metanormneg m =
  match m with
  | MAp(MAp(MGround(Imp),m1),MGround(False)) -> m1
  | _ -> MAp(MAp(MGround(Imp),m),MGround(False))

let rec meta_copy m evarassoc =
  begin
    match m with
    | MGround(m1) -> m
    | MVar((e,x),sigma) ->
	let sigma1 = List.map (fun m -> meta_copy m evarassoc) sigma in
	begin
	  match (!x) with
	  | None ->
	      begin
		try
		  let (_,y) = List.find (fun (e1,_) -> (e == e1)) evarassoc in
		  MVar(y,sigma1)
		with
		| Not_found ->
		    MVar((e,x),sigma1)
	      end
	  | Some m1 ->
	      let m1c = meta_copy m1 evarassoc in
	      metanorm (meta_simulsubst_meta m1c sigma1)
	end
    | MLam(a1,m1) ->
	let m1c = meta_copy m1 evarassoc in
	MLam(a1,m1c)
    | MAp(m1,m2) ->
	let m1c = meta_copy m1 evarassoc in
	let m2c = meta_copy m2 evarassoc in
	MAp(m1c,m2c)
  end

let rec distinct_bvar_list_p_rec ml dl =
  match ml with
  | [] -> true
  | ((MGround(DB(i,_)))::mr) when (not (List.mem i dl)) -> distinct_bvar_list_p_rec mr (i::dl)
  | _ -> false

let distinct_bvar_list_p ml =
  distinct_bvar_list_p_rec ml []

let mvar_p m =
  match m with
  | MVar _ -> true
  | _ -> false

let rec pattern_p m =
  match m with
  | MGround(m1) -> true
  | MVar((e,x),ml) ->
      begin
	match (!x) with
	| None -> distinct_bvar_list_p ml
	| Some n ->
	    pattern_p (meta_simulsubst_meta n ml) (*** Mar 2012 - finally wrote this case ***)
      end
  | MLam(a,m1) -> pattern_p m1
  | MAp(m1,m2) -> pattern_p m1 && pattern_p m2

let rec meta_head_spine_rec m args =
  match m with
  | MAp(m1,m2) -> meta_head_spine_rec m1 (m2::args)
  | MGround(m1) -> (m1,tpof m1,args)
  | _ -> raise (Failure "Unexpected case in meta_head_spine_rec")

let meta_head_spine m =
  meta_head_spine_rec m []

let rec occurs_check_p (e,x) m =
  match m with
  | MGround(_) -> false
  | MVar((e',y),_) when (e = e') -> true
  | MVar((e',y),ml) -> occurs_check_list_p (e,x) ml
  | MLam(a,m1) -> occurs_check_p (e,x) m1
  | MAp(m1,m2) -> occurs_check_p (e,x) m1 || occurs_check_p (e,x) m2
and occurs_check_list_p x ml =
  match ml with
  | (m::mr) -> occurs_check_p x m || occurs_check_list_p x mr
  | [] -> false

let rec pattern_invert_db ml n j i k =
  match ml with
  | ((MGround(DB(i2,a)))::mr) when i = i2 -> MGround(DB(j+k,a))
  | (_::mr) -> pattern_invert_db mr n j i (k + 1)
  | [] -> raise PatternMatchFailed

let rec pattern_invert x ml n j =
  match n with
  | DB(i,_) ->
      if (i < j) then
	MGround(n)
      else
	pattern_invert_db ml n j (i - j) 0
  | Lam(a,n1) -> MLam(a,pattern_invert x ml n1 (j + 1))
  | Ap(n1,n2) -> MAp(pattern_invert x ml n1 j,pattern_invert x ml n2 j)
  | _ -> MGround(n)
and pattern_list_invert x ml nl j =
  match nl with
  | (n::nr) -> (pattern_invert x ml n j::pattern_list_invert x ml nr j)
  | [] -> []

let rec pattern_match_rec congforms dl cl =
  match dl with
  | [] -> cl
  | (d::dr) -> pattern_match_rec_1 congforms d dr cl
and pattern_match_rec_1 congforms d dl cl =
  match d with
  | (gamma,m1,n1,Ar(a,b)) ->
      pattern_match_rec_1 congforms (a::gamma,gen_mlam_body a m1,gen_lam_body a n1,b) dl cl
  | (gamma,MVar((e,x),ml),n,b) ->
      begin
	match (!x) with
	| None ->
	    begin
	      if (distinct_bvar_list_p ml) then
		let ni = pattern_invert (e,x) ml n 0 in
		begin
		  x := Some ni;
		  let clr = ref [] in
		  let dlr = ref dl in
		  List.iter
		    (fun ((gammac,mc,nc,b) as c) ->
		      let (mc1,mb) = metanorm1 mc in
		      if mb then
			let mcn = metanorm mc1 in
			dlr := ((gammac,mcn,nc,b)::!dlr)
		      else
			clr := (c::(!clr))
		    )
		    cl;
		  pattern_match_rec congforms (!dlr) (!clr)
		end
	      else
		begin
		  pattern_match_rec congforms dl (d::cl)
		end
	    end
	| Some(mx) -> pattern_match_rec_1 congforms (gamma,metanorm (meta_simulsubst_meta mx ml),n,b) dl cl
      end
  | (gamma,m,n,b) ->
      let (mh,mhtp,ml) = meta_head_spine m in
      try
        let (nh,nl) = head_spine n in
	if (mh = nh) then
	  pattern_match_rec_spine congforms gamma mhtp ml nl dl cl
	else
	  raise PatternMatchFailed
      with PatternMatchFailed -> (*** try harder using congruent forms of n ***)
	pattern_match_rec_2 congforms gamma mh mhtp ml (congforms n) b dl cl
and pattern_match_rec_2 congforms gamma mh mhtp ml ncl b dl cl =
  match ncl with
  | [] -> raise PatternMatchFailed
  | (n::ncr) ->
      try
	let (nh,nl) = head_spine n in
	if (mh = nh) then
	  pattern_match_rec_spine congforms gamma mhtp ml nl dl cl
	else
	  raise PatternMatchFailed
      with PatternMatchFailed -> pattern_match_rec_2 congforms gamma mh mhtp ml ncr b dl cl
and pattern_match_rec_spine congforms gamma tp ml nl dl cl =
  match (tp,ml,nl) with
  | (_,[],[]) -> pattern_match_rec congforms dl cl
  | (Ar(a,b),(m::ml),(n::nl)) -> pattern_match_rec_spine congforms gamma b ml nl ((gamma,m,n,a)::dl) cl
  | _ -> raise PatternMatchFailed
      
let pattern_match congforms dl = pattern_match_rec congforms dl []

let mem_evar (d,a) l =
  try
    ignore(List.find (fun (e,x) -> e = d) l);
    true
  with Not_found -> false

let rec update_strict xl m =
  match m with
  | MGround(_) -> xl
  | MVar(x,sigma) ->
      if (mem_evar x xl) then
	xl
      else if (distinct_bvar_list_p sigma) then
	(x::xl)
      else
	xl
  | MAp(m1,m2) -> update_strict (update_strict xl m1) m2
  | MLam(a1,m1) -> update_strict xl m1

(** HOUNIF1 - Mar 2012 - Chad **)
type udpair = ctx * metatrm * metatrm * stp
type cdpairs = int * evar list * metatrm list * (ctx * metatrm * metatrm * stp) list
type frpairstp = (ctx * (metatrm * (int * metatrm option ref) * metatrm list) *
         (metatrm * Syntax.trm * Syntax.stp * metatrm list) * stp) list
type ffpairstp = (ctx * (metatrm * (int * metatrm option ref) * metatrm list) *
		    (metatrm * (int * metatrm option ref) * metatrm list) * stp) list

exception SimplFailed
exception SimplElim of evar * metatrm * udpair list * frpairstp * ffpairstp

let verbosity : int ref = ref 0;;

let rec id_subst_2 i gamma =
  match gamma with
  | (a::gammar) -> ((DB(i,a))::(id_subst_2 (i + 1) gammar))
  | [] -> []

let meta_to_ground_2 m tau =
   norm emptyd (meta_to_ground_rec m tau)

let rec spine_dpairs gam a args1 args2 dpairs =
  match (a,args1,args2) with
  | (Ar(a,b),(m1::args1),(m2::args2)) -> spine_dpairs gam b args1 args2 ((gam,m1,m2,a)::dpairs)
  | (_,[],[]) -> dpairs
  | _ -> raise (Failure "Unexpected case in spine_dpairs")

let rec metapattern_invert_db ml n j i k =
  match ml with
  | ((MGround(DB(i2,a)))::mr) when i = i2 -> MGround(DB(j+k,a))
  | (_::mr) -> pattern_invert_db mr n j i (k + 1)
  | [] -> raise PatternMatchFailed

let rec metapattern_invert (e,x) ml n j =
  match n with
  | MGround(DB(i,_)) ->
      if (i < j) then
	n
      else
	metapattern_invert_db ml n j (i - j) 0
  | MGround(_) -> n
  | MLam(a,n1) -> MLam(a,metapattern_invert (e,x) ml n1 (j + 1))
  | MAp(n1,n2) -> MAp(metapattern_invert (e,x) ml n1 j,metapattern_invert (e,x) ml n2 j)
  | MVar((d,y),sl) when e = d -> raise PatternMatchFailed (*** occurs check ***)
  | MVar((d,y),sl) -> MVar((d,y),metapattern_list_invert (e,x) ml sl j)
and metapattern_list_invert x ml nl j =
  match nl with
  | (n::nr) -> (metapattern_invert x ml n j::metapattern_list_invert x ml nr j)
  | [] -> []

let pack_into_dpairs dpairs frpairs ffpairs =
  (List.append (List.map (fun (gam,m1,m2,a) -> (gam,metanorm m1,metanorm m2,a)) dpairs)
     (List.append
	(List.map (fun (gam,(m1,_,_),(m2,_,_,_),a) -> (gam,metanorm m1,metanorm m2,a)) frpairs)
	(List.map (fun (gam,(m1,_,_),(m2,_,_),a) -> (gam,metanorm m1,metanorm m2,a)) ffpairs)))

let rec simpl_dpairs (dpairs:udpair list) xl frpairs ffpairs =
  if (!verbosity > 5) then
    begin
      Printf.printf "simpl_dpairs\n";
      List.iter (fun (gam,m1,m2,b) -> Printf.printf ". %s\nm1 = %s\nm2 = %s\n" (stp_str b) (metatrm_str m1) (metatrm_str m2)) dpairs;
    end;
  match dpairs with
  | (gam,m1,m2,Ar(a,b))::dpairs ->
      if (!verbosity > 5) then begin Printf.printf "simpl_dpairs 0\n" end;
      simpl_dpairs ((a::gam,gen_mlam_body a m1,gen_mlam_body a m2,b)::dpairs) xl frpairs ffpairs
  | (gam,((MVar(x,sl1)) as m1),m2,b)::dpairs when distinct_bvar_list_p sl1 -> (*** pattern ***)
      begin
	try
	  let m2' = metapattern_invert x sl1 m2 0 in
	  raise (SimplElim(x,m2',dpairs,frpairs,ffpairs))
	with PatternMatchFailed ->
	  simpl_dpairs_2 dpairs xl frpairs ffpairs gam m1 m2 b
      end
  | (gam,m1,((MVar(x,sl2)) as m2),b)::dpairs when distinct_bvar_list_p sl2 -> (*** pattern ***)
      begin
	try
	  let m1' = metapattern_invert x sl2 m1 0 in
	  raise (SimplElim(x,m1',dpairs,frpairs,ffpairs))
	with PatternMatchFailed ->
	  simpl_dpairs_2 dpairs xl frpairs ffpairs gam m1 m2 b
      end
  | (gam,m1,m2,b)::dpairs ->
      simpl_dpairs_2 dpairs xl frpairs ffpairs gam m1 m2 b
  | [] -> (frpairs,ffpairs)
and simpl_dpairs_2 dpairs xl frpairs ffpairs gam m1 m2 b =
  if (!verbosity > 5) then begin Printf.printf "simpl_dpairs_2\n" end;
  begin
    match (m1,m2) with
    | (MVar(x1,sigma1),MVar(x2,sigma2)) ->
	simpl_dpairs dpairs xl frpairs ((gam,(m1,x1,sigma1),(m2,x2,sigma2),b)::ffpairs)
    | (_,MVar(x2,sigma2)) ->
	begin
	  match (meta_head_spine m1) with
	  | (h1,h1tp,args1) ->
	      simpl_dpairs dpairs xl ((gam,(m2,x2,sigma2),(m1,h1,h1tp,args1),b)::frpairs) ffpairs
	end
    | (MVar(x1,sigma1),_) ->
	begin
	  match (meta_head_spine m2) with
	  | (h2,h2tp,args2) ->
	      simpl_dpairs dpairs xl ((gam,(m1,x1,sigma1),(m2,h2,h2tp,args2),b)::frpairs) ffpairs
	end
    | (_,MAp(MAp(MGround(Imp),MVar(x2,sigma2)),MGround(False))) when not (mneg_p m1) && !hounif1negflex ->
	simpl_dpairs dpairs xl ((gam,(MVar(x2,sigma2),x2,sigma2),(mneg m1,Imp,Ar(Prop,Ar(Prop,Prop)),[m1;MGround(False)]),b)::frpairs) ffpairs
    | (MAp(MAp(MGround(Imp),MVar(x1,sigma1)),MGround(False)),_) when not (mneg_p m2) && !hounif1negflex ->
	simpl_dpairs dpairs xl ((gam,(MVar(x1,sigma1),x1,sigma1),(mneg m2,Imp,Ar(Prop,Ar(Prop,Prop)),[m2;MGround(False)]),b)::frpairs) ffpairs
    | _ ->
	begin
	  match (meta_head_spine m1,meta_head_spine m2) with
	  | ((h1,h1tp,args1),(h2,h2tp,args2)) ->
	      if (h1 = h2) then
		simpl_dpairs
		  (spine_dpairs gam h1tp args1 args2 dpairs)
		  xl
		  frpairs ffpairs
	      else
		raise SimplFailed
	end
  end

let defaultinsts : (int,metatrm) Hashtbl.t = Hashtbl.create 100
let evargamma : (int,ctx) Hashtbl.t = Hashtbl.create 100
let evarprojs : (int,(int * stp * stp list * stp) list) Hashtbl.t = Hashtbl.create 100

let defaultmeta a =
  match a with
  | Prop -> MGround(False)
  | _ -> MAp(MGround(Choice(a)),MLam(a,MGround(False)))

let rec new_evar_w_default gamma a g =
  match a with
  | Ar(a1,a2) ->
      let (x,xs) = new_evar_w_default (a1::gamma) a2 (fun n -> MLam(a1,g n)) in
      (x,MLam(a1,xs))
  | _ ->
      let x = (!evarcount,ref None) in
      let s = id_subst 0 gamma in
      Hashtbl.add defaultinsts !evarcount (g (defaultmeta a));
      Hashtbl.add evargamma !evarcount gamma;
      let i = ref 0 in
      let projl = ref [] in
      List.iter (fun b ->
	let (argtps,rtp) = argtps_rtp b in
	if (rtp = a) then
	  begin
	    if (!verbosity > 20) then Printf.printf "%d is a possible projection for %s\n" !i (string_of_evar x);
	    projl := (!i,b,argtps,a)::!projl;
	  end;
	incr i;
	)
	gamma;
      Hashtbl.add evarprojs !evarcount !projl;
      incr evarcount;
      (x,MVar(x,s))

let rec force_meta_to_ground_rec m tau =
  if (!verbosity > 5) then begin Printf.printf "force_meta_to_ground_rec m %s\n" (metatrm_str m) end;
   match m with
   | MGround(m1) -> simulsubst m1 tau
   | MVar((e,x),sigma) ->
       begin
	 match (!x) with
	 | None ->
	     let n = Hashtbl.find defaultinsts e in
	     x := Some n;
	     force_meta_to_ground_rec n (List.map (fun p -> force_meta_to_ground_rec p tau) sigma)
	 | Some n ->
	     force_meta_to_ground_rec n (List.map (fun p -> force_meta_to_ground_rec p tau) sigma)
       end
   | MLam(a,m1) -> Lam(a,force_meta_to_ground_rec m1 ((DB(0,a))::(List.map (fun n -> shift n 0 1) tau)))
   | MAp(m1,m2) -> Ap(force_meta_to_ground_rec m1 tau,force_meta_to_ground_rec m2 tau)

let rec imitate gam gam' sigma h a args =
  match (a,args) with
  | (Ar(a1,a2),(arg1::argr)) ->
      let (z,zsub) = new_evar_w_default gam a1 (fun x -> x) in
      let (g,zl,dpairs2) = imitate gam gam' sigma (MAp(h,zsub)) a2 argr in
      (g,(z::zl),((gam',(meta_simulsubst_meta zsub sigma),arg1,a1)::dpairs2))
  | (_,[]) -> (h,[],[])
  | _ -> raise (Failure "Unexpected case in imitate")

let rec project gam h a =
  match a with
  | Ar(a1,a2) ->
      let (z,zsub) = new_evar_w_default gam a1 (fun x -> x) in
      let (g,zl) = project gam (MAp(h,zsub)) a2 in
      (g,(z::zl))
  | _ -> (h,[])

let rec union_evars xl yl =
  match xl with
  | (x::xr) ->
      begin
	try
	  ignore (List.find (fun x' -> x' == x) yl);
	  union_evars xr yl
	with
	  Not_found -> union_evars xr (x::yl)
      end
  | [] -> yl

let rec union_metatrms ml nl =
  match ml with
  | (m::mr) ->
      begin
	if (List.mem m nl) then
	  union_metatrms mr nl
	else
	  union_metatrms mr (m::nl)
      end
  | [] -> nl

let mate_dpairs : (int * cdpairs) list ref = ref [];;
let metaflexposatoms : (evar list * metatrm list * metatrm) list ref = ref [];;
let metaflexnegatoms : (evar list * metatrm list * metatrm) list ref = ref [];;
let metaposatoms : (evar list * metatrm list * metatrm) list ref = ref [];;
let metanegatoms : (evar list * metatrm list * metatrm) list ref = ref [];;

(*** copied from match.ml to modify for cong clos END ***)

let collecttypingaxioms : bool ref = ref false
type mizartypingaxiom =
  | MizarTypeHierarchyStrict of string * string list * int * trm list * trm list
  | MizarTypeHierarchy of string * string list * int * trm list * trm list
  | MizarFunctionType of string * string list * int * trm list * trm list
  | MizarTypingGeneric of string * int * trm list * trm list
  | MizarRedefinition of string * int * trm list * string * trm list * trm
let typingaxiomslist : mizartypingaxiom list ref = ref []

let mizar_pattern_table : (string,int * trm list * stp * int list * trm * trm) Hashtbl.t = Hashtbl.create 1000
let mizar_pattern_list : (string * int * trm list * stp * int list * trm * trm) list ref = ref []

let mizar_symbol_order_pred : (string, string) Hashtbl.t ref = ref (Hashtbl.create 1000)
let mizar_symbol_order_succ : (string, string) Hashtbl.t = Hashtbl.create 1000
let mizar_symbol_order_mem : (string * string,unit) Hashtbl.t = Hashtbl.create 1000

let mizar_name_tp : (string,stp) Hashtbl.t ref = ref (Hashtbl.create 1000)

let mizar_func_types : (string,string * int * trm list * trm list) Hashtbl.t = Hashtbl.create 20000
let mizar_typing : (string,string * int * trm * trm list * trm list) Hashtbl.t = Hashtbl.create 5000
let mizar_typing_gen_forward : (string,string * int * trm * trm list * trm list) Hashtbl.t = Hashtbl.create 5000
let mizar_typing_gen_backward : (string,string * int * trm * trm list * trm) Hashtbl.t = Hashtbl.create 5000
let mizar_redef : (string,string * int * trm list * trm list * trm) Hashtbl.t = Hashtbl.create 5000

let name_prefix x =
  let l = String.length x in
  let b = ref false in
  let y = ref "" in
  let z = ref (-1) in
  try
    for i = 0 to l - 1 do
      let c = x.[i] in
      let cc = Char.code c in
      if !b then
	if cc >= 48 && cc <= 57 then
	  z := !z * 10 + cc - 48
	else
	  raise Exit
      else
	if cc >= 48 && cc <= 57 then
	  (b := true; z := cc - 48)
	else if c = '_' then
	  raise Exit
	else
	  y := Printf.sprintf "%s%c" !y c
    done;
    (!y,!z)
  with Exit ->
    (!y,!z)

let add_mizar_symbol_order x y =
  if not (Hashtbl.mem mizar_symbol_order_mem (x,y)) then
    let a u v =
      Hashtbl.add mizar_symbol_order_mem (u,v) ();
      Hashtbl.add mizar_symbol_order_succ u v;
      Hashtbl.add !mizar_symbol_order_pred v u;
    in
    let b u v =
      if not (Hashtbl.mem mizar_symbol_order_mem (x,y)) then
	begin
	  try
	    a u v
	  with Not_found ->
	    a u v
	end
    in
    let zl = Hashtbl.find_all mizar_symbol_order_succ y in
    let wl = Hashtbl.find_all !mizar_symbol_order_pred x in
    a x y;
    List.iter (fun z -> b x z) zl;
    List.iter
      (fun w ->
	b w y;
	List.iter (fun z -> b w z) zl)
      wl

let rec extract_conjuncts m pol =
  match m with
  | Ap(Ap(And,m1),m2) when pol -> extract_conjuncts m1 true @ extract_conjuncts m2 true
  | Ap(Ap(Or,m1),m2) when not pol -> extract_conjuncts m1 false @ extract_conjuncts m2 false
  | Ap(Neg,m1) -> extract_conjuncts m1 (not pol)
  | Ap(Ap(Imp,m1),False) -> extract_conjuncts m1 (not pol)
  | Ap(Ap(Imp,m1),m2) when not pol -> extract_conjuncts m1 true @ extract_conjuncts m2 false
  | False when not pol -> []
  | True when pol -> []
  | _ -> if pol then [m] else [neg m]

let rec andlist f l =
  match l with
  | x::r -> andlist f r && f x
  | [] -> true

let names_in_mizar_type_assumptions : string list ref = ref []
let mizar_hook_names : string list ref = ref []

let rec fofunstp_p a =
  match a with
  | Base("$i") -> true
  | Ar(Base("$i"),b) -> fofunstp_p b
  | _ -> false

let rec foterm_p m =
  let (h,s) = head_spine m in
  match h with
  | DB(i,Base("$i")) when s = [] -> true
  | Name(f,a) -> fofunstp_p a && foterm_spine_p s
  | _ -> false
and foterm_spine_p s =
  match s with
  | [] -> true
  | m::r -> foterm_p m && foterm_spine_p r

let rec mizar_type_lit_genoa_p c m =
  match m with
  | Ap(Ap(Imp,m1),False) -> mizar_type_lit_genoa_p c m1
  | _ ->
      let (h,s) = head_spine m in
      match h with
      | Name(p,_) ->
	  if foterm_spine_p s then
	    begin
	      names_in_mizar_type_assumptions := p::!names_in_mizar_type_assumptions;
	      let (pr,_) = name_prefix p in
	      if pr = "m" || pr = "l" || pr = "v" then c s else false
	    end
	  else
	    false
      | _ -> false

let rec mizar_type_lit_genoc_p c m =
  match m with
  | Ap(Ap(Imp,m1),False) -> mizar_type_lit_genoc_p c m1
  | _ ->
      let (h,s) = head_spine m in
      match h with
      | Name(p,_) ->
	  if foterm_spine_p s then
	    let (pr,_) = name_prefix p in
	    if (pr = "m" || pr = "l" || pr = "v") && c s then
	      if List.exists
		  (fun x -> x = p || Hashtbl.mem mizar_symbol_order_mem (p,x))
		  !names_in_mizar_type_assumptions
	      then
		begin
		  false
		end
	      else
		begin
		  if not (List.mem p !mizar_hook_names) then mizar_hook_names := p::!mizar_hook_names;
		  List.iter (fun x -> add_mizar_symbol_order x p) !names_in_mizar_type_assumptions;
		  true
		end
	    else
	      false
	  else
	    false
      | _ -> false

let rec mizar_type_lit_gen_p c m =
  match m with
  | Ap(Ap(Imp,m1),False) -> mizar_type_lit_gen_p c m1
  | _ ->
      let (h,s) = head_spine m in
      match h with
      | Name(p,_) ->
	  if foterm_spine_p s then
	    let (pr,_) = name_prefix p in
	    if pr = "m" || pr = "l" || pr = "v" then c s else false
	  else
	    false
      | _ -> false

let mizar_type_of_var_p m =
  mizar_type_lit_gen_p
    (fun s ->
      match s with
      | (DB(_,_)::_) -> true
      | _ -> false)
   m

let mizar_type_of_varoa_p m =
  mizar_type_lit_genoa_p
    (fun s ->
      match s with
      | (DB(_,_)::_) -> true
      | _ -> false)
   m

let mizar_type_of_varoc_p m =
  mizar_type_lit_genoc_p
    (fun s ->
      match s with
      | (DB(_,_)::_) -> true
      | _ -> false)
   m

let mizar_type_of_vars_p m =
  mizar_type_lit_gen_p
    (fun s -> andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s)
    m

let mizar_type_of_varsoa_p m =
  mizar_type_lit_genoa_p
    (fun s -> andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s)
    m

let mizar_type_of_result_p m =
  mizar_type_lit_gen_p
    (fun s ->
      match s with
      | (a::_) ->
	  begin
	    match head_spine a with
	    | (Name(_,_),l) -> andlist (fun x -> match x with DB(_,_) -> true | _ -> false) l
	    | _ -> false
	  end
      | [] -> false)
    m

let rec spine_countdown_p l i =
  match l with
  | DB(j,Base("$i"))::r when j = i -> spine_countdown_p r (i-1)
  | [] -> i = (-1)
  | _ -> false

let mizar_type_of_resultoc_p i m =
  if mizar_type_lit_genoa_p (fun _ -> true) m then (*** just to collect the name of type conclusion type as if it's an assumption for dependency checking ***)
    mizar_type_lit_gen_p (*** call gen_ and not genoc_ here because we're not checking that the type in the conclusion comes later than the types in the assumption; but are checking if the type in the conclusion comes before the function ***)
      (fun s ->
	match s with
	| (a::_) ->
	    begin
	      match head_spine a with
	      | (Name(p,_),l) -> (*** insist that l is of the form [DB(i-1,Base("$i"));...;DB(0,Base("$i")] ***)
		  if spine_countdown_p l (i-1) then
		    begin
		      if List.exists
			  (fun x -> x = p || Hashtbl.mem mizar_symbol_order_mem (p,x))
			  !names_in_mizar_type_assumptions
		      then
			false
		      else
			begin
			  if not (List.mem p !mizar_hook_names) then mizar_hook_names := p::!mizar_hook_names;
			  List.iter (fun x -> add_mizar_symbol_order x p) !names_in_mizar_type_assumptions;
			  true
			end
		    end
		  else
		    false
		  (*** More general variant that would require permuting / matching
		  let vo = Array.create i false in (*** check that every var occurs as an arg of the func ***)
		  if andlist (fun x -> match x with DB(j,_) when j < i && j >= 0 -> (vo.(j) <- true; true) | _ -> false) l then
		    let missingvar = ref false in
		    begin
		      for j = 0 to i-1 do
			if not vo.(j) then missingvar := true
		      done;
		      if !missingvar then
			false
		      else if List.exists
			  (fun x -> x = p || Hashtbl.mem mizar_symbol_order_mem (p,x))
			  !names_in_mizar_type_assumptions
		      then
			false
		      else
			begin
			  if not (List.mem p !mizar_hook_names) then mizar_hook_names := p::!mizar_hook_names;
			  List.iter (fun x -> add_mizar_symbol_order x p) !names_in_mizar_type_assumptions;
			  true
			end
		    end
		  else
		     false
		     ***)
	      | _ -> false
	    end
	| [] -> false)
      m
  else
    false
      
let rec mizar_type_fact1_p m i pl =
  match m with
  | Ap(Forall(Base("$i")),Lam(_,m1)) -> mizar_type_fact1b_p m1 (i+1) (List.map (fun p -> shift p 0 1) pl)
  | _ ->
      let ml = extract_conjuncts m true in
      if andlist (mizar_type_lit_gen_p (fun _ -> true)) ml then
	(i,pl,ml)
      else
	raise Not_found
and mizar_type_fact1b_p m i pl =
  match m with
  | Ap(Ap(Imp,m1),False) ->
      let ml = extract_conjuncts m1 false in
      if andlist (mizar_type_lit_gen_p (fun _ -> true)) ml then
	(i,pl,ml)
      else
	raise Not_found
  | Ap(Ap(Imp,m1),m2) ->
      let ql = extract_conjuncts m1 true in
      if andlist mizar_type_of_var_p ql then
	mizar_type_fact1b_p m2 i (ql @ pl)
      else
	raise Not_found
  | _ -> mizar_type_fact1_p m i pl

let rec radixof_n_p m n =
  match m with
  | Ap(Name(x,_),DB(0,_)) ->
      begin
	let (pr,_) = name_prefix x in
	pr = "m" || pr = "l"
      end
  | Ap(m1,DB(k,_)) when k = n -> radixof_n_p m1 (n+1)
  | _ -> false

let radixof_p m =
  match m with
  | Ap(Name(x,_),DB(0,_)) ->
      begin
	let (pr,_) = name_prefix x in
	pr = "m" || pr = "l"
      end
  | Ap(m1,DB(n,_)) when n = 1 -> radixof_n_p m1 (n+1)
  | _ -> false

let rec moreforalls_p m =
  match m with
  | Ap(Forall(_),_) -> true
  | Ap(Ap(Imp,m1),m2) -> moreforalls_p m2
  | _ -> false

let rec mizar_type_fact2_p m =
  match m with
  | Ap(Forall(Base("$i")),Lam(_,Ap(Ap(Imp,m1),m2))) when radixof_p m1 && not (moreforalls_p m2) ->
      let ml = extract_conjuncts m2 true in
      andlist mizar_type_of_var_p ml
  | Ap(Forall(Base("$i")),Lam(_,m1)) -> mizar_type_fact2b_p m1
  | _ -> false
and mizar_type_fact2b_p m =
  match m with
  | Ap(Ap(Imp,m1),m2) -> andlist mizar_type_of_var_p (extract_conjuncts m1 true) && mizar_type_fact2b_p m2
  | _ -> mizar_type_fact2_p m

let rec mizar_type_fact2o_p m =
  match m with
  | Ap(Forall(Base("$i")),Lam(_,Ap(Ap(Imp,m1),m2))) when radixof_p m1 && not (moreforalls_p m2) ->
      let ml = extract_conjuncts m2 true in
      if andlist mizar_type_of_varoa_p ml then
	begin
	  let (h,_) = head_spine m1 in (*** TODO: I should also check that the m1 has all the DB's of the fact so that matching will instantiate all vars. ***)
	  match h with
	  | Name(p,_) ->
	      if List.exists
		  (fun x -> x = p || Hashtbl.mem mizar_symbol_order_mem (p,x))
		  !names_in_mizar_type_assumptions
	      then
		begin
		  false
		end
	      else
		begin
		  if not (List.mem p !mizar_hook_names) then mizar_hook_names := p::!mizar_hook_names;
		  List.iter (fun x -> add_mizar_symbol_order x p) !names_in_mizar_type_assumptions;
		  true
		end
	  | _ -> false
	end
      else
	false
  | Ap(Forall(Base("$i")),Lam(_,m1)) -> mizar_type_fact2ob_p m1
  | _ -> false
and mizar_type_fact2ob_p m =
  match m with
  | Ap(Ap(Imp,m1),m2) -> andlist mizar_type_of_varoa_p (extract_conjuncts m1 true) && mizar_type_fact2ob_p m2
  | _ -> mizar_type_fact2o_p m
	
let rec mizar_type_fact3_p m =
  match m with
  | Ap(Forall(Base("$i")),Lam(_,Ap(Ap(Imp,m1),m2))) when radixof_p m1 && not (moreforalls_p m2) ->
      let ml = extract_conjuncts m2 true in
      andlist mizar_type_of_var_p ml
  | Ap(Forall(Base("$i")),Lam(_,m1)) -> mizar_type_fact3b_p m1
  | _ -> false
and mizar_type_fact3b_p m =
  match m with
  | Ap(Ap(Imp,m1),m2) -> andlist mizar_type_of_vars_p (extract_conjuncts m1 true) && mizar_type_fact3b_p m2
  | _ -> mizar_type_fact3_p m

let rec mizar_type_fact3o_p m =
  match m with
  | Ap(Forall(Base("$i")),Lam(_,Ap(Ap(Imp,m1),m2))) when radixof_p m1 && not (moreforalls_p m2) ->
      let ml = extract_conjuncts m2 true in
      if andlist mizar_type_of_varoa_p ml then
	begin
	  let (h,_) = head_spine m1 in (*** TODO: I should also check that the m1 has all the DB's of the fact so that matching will instantiate all vars. ***)
	  match h with
	  | Name(p,_) ->
	      if List.exists
		  (fun x -> x = p || Hashtbl.mem mizar_symbol_order_mem (p,x))
		  !names_in_mizar_type_assumptions
	      then
		begin
		  false
		end
	      else
		begin
		  if not (List.mem p !mizar_hook_names) then mizar_hook_names := p::!mizar_hook_names;
		  List.iter (fun x -> add_mizar_symbol_order x p) !names_in_mizar_type_assumptions;
		  true
		end
	  | _ -> false
	end
      else
	false
  | Ap(Forall(Base("$i")),Lam(_,m1)) -> mizar_type_fact3ob_p m1
  | _ -> false
and mizar_type_fact3ob_p m =
  match m with
  | Ap(Ap(Imp,m1),m2) -> andlist mizar_type_of_varsoa_p (extract_conjuncts m1 true) && mizar_type_fact3ob_p m2
  | _ -> mizar_type_fact3o_p m
	
let rec mizar_type_fact4_p m =
  match m with
  | Ap(Forall(Base("$i")),Lam(_,m1)) -> mizar_type_fact4b_p m1
  | _ ->
      let ml = extract_conjuncts m true in
      andlist mizar_type_of_result_p ml
and mizar_type_fact4b_p m =
  match m with
  | Ap(Ap(Imp,m1),False) ->
      let ml = extract_conjuncts m1 false in
      andlist mizar_type_of_result_p ml
  | Ap(Ap(Imp,m1),m2) -> andlist mizar_type_of_var_p (extract_conjuncts m1 true) && mizar_type_fact4b_p m2
  | _ -> mizar_type_fact4_p m

let rec mizar_type_fact4oa_p i m =
  match m with
  | Ap(Forall(Base("$i")),Lam(_,m1)) -> mizar_type_fact4ob_p (i+1) m1
  | _ ->
      let ml = extract_conjuncts m true in
      andlist (mizar_type_of_resultoc_p i) ml
and mizar_type_fact4ob_p i m =
  match m with
  | Ap(Ap(Imp,m1),False) ->
      let ml = extract_conjuncts m1 false in
      andlist (mizar_type_of_resultoc_p i) ml
  | Ap(Ap(Imp,m1),m2) -> andlist mizar_type_of_varoa_p (extract_conjuncts m1 true) && mizar_type_fact4ob_p i m2
  | _ -> mizar_type_fact4oa_p i m

let mizar_type_fact4o_p m = mizar_type_fact4oa_p 0 m
	
let mizar_abbrev_eqn m =
  match m with
  | Ap(Ap(Eq(_),m1),m2) ->
      begin
	let (h1,s1) = head_spine m1 in
	let (h2,s2) = head_spine m2 in
	match (h1,h2) with
	| (Name(f1,_),Name(f2,_)) when not (f1 = f2) ->
	    if andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s1
		&&
	      andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s2
		&&
	      List.for_all (fun x -> List.mem x s1) s2
	    then
	      (f1,s1,m2)
	    else
	      raise Not_found
	| (Name(f1,_),Eq(_)) ->
	    if andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s1
		&&
	      andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s2
		&&
	      List.for_all (fun x -> List.mem x s1) s2
	    then
	      (f1,s1,m2)
	    else
	      raise Not_found
	| _ -> raise Not_found
      end
  | _ -> raise Not_found

let mizar_abbrev_eqn_o m =
  match m with
  | Ap(Ap(Eq(_),m1),m2) ->
      begin
	let (h1,s1) = head_spine m1 in
	let (h2,s2) = head_spine m2 in
	match (h1,h2) with
	| (Name(f1,_),Name(f2,_)) when not (f1 = f2) ->
	    if andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s1
		&&
	      andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s2
		&&
	      List.for_all (fun x -> List.mem x s1) s2
	    then
	      begin
		names_in_mizar_type_assumptions := f2::!names_in_mizar_type_assumptions;
		if List.exists
		    (fun x -> x = f1 || Hashtbl.mem mizar_symbol_order_mem (f1,x))
		    !names_in_mizar_type_assumptions
		then
		  raise Not_found
		else
		  begin
		    List.iter (fun x -> add_mizar_symbol_order x f1) !names_in_mizar_type_assumptions;
		    (f1,s1,m2)
		  end
	      end
	    else
	      raise Not_found
	| (Name(f1,_),Eq(_)) ->
	    if andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s1
		&&
	      andlist (fun x -> match x with DB(_,_) -> true | _ -> false) s2
		&&
	      List.for_all (fun x -> List.mem x s1) s2
	    then
	      if List.exists
		  (fun x -> x = f1 || Hashtbl.mem mizar_symbol_order_mem (f1,x))
		  !names_in_mizar_type_assumptions
	      then
		raise Not_found
	      else
		begin
		  List.iter (fun x -> add_mizar_symbol_order x f1) !names_in_mizar_type_assumptions;
		  (f1,s1,m2)
		end
	    else
	      raise Not_found
	| _ -> raise Not_found
      end
  | _ -> raise Not_found

let rec mizar_type_fact5_p m i pl =
  match m with
  | Ap(Forall(Base("$i")),Lam(_,m1)) -> mizar_type_fact5b_p m1 (i+1) (List.map (fun p -> shift p 0 1) pl)
  | _ ->
      let (f1,s1,m2) = mizar_abbrev_eqn m in
      (i,pl,f1,s1,m2)
and mizar_type_fact5b_p m i pl =
  match m with
  | Ap(Ap(Imp,m1),False) -> raise Not_found
  | Ap(Ap(Imp,m1),m2) ->
      let ql = extract_conjuncts m1 true in
      if andlist mizar_type_of_var_p ql then
	mizar_type_fact5b_p m2 i (pl @ ql)
      else
	raise Not_found
  | _ -> mizar_type_fact5_p m i pl

let rec mizar_type_fact5o_p m i pl =
  match m with
  | Ap(Forall(Base("$i")),Lam(_,m1)) -> mizar_type_fact5ob_p m1 (i+1) (List.map (fun p -> shift p 0 1) pl)
  | _ ->
      let (f1,s1,m2) = mizar_abbrev_eqn_o m in
      (i,pl,f1,s1,m2)
and mizar_type_fact5ob_p m i pl =
  match m with
  | Ap(Ap(Imp,m1),False) -> raise Not_found
  | Ap(Ap(Imp,m1),m2) ->
      let ql = extract_conjuncts m1 true in
      if andlist mizar_type_of_varoa_p ql then
	mizar_type_fact5ob_p m2 i (pl @ ql)
      else
	raise Not_found
  | _ -> mizar_type_fact5o_p m i pl
	
let rec mizar_pattern m i pl =
(*  Printf.printf "mizar_pattern %d %s\n" i (trm_str m); flush stdout; *)
  match m with
  | Ap(Forall(Base("$i")),Lam(_,m1)) ->
      mizar_pattern m1 (i+1) (List.map (fun p -> shift p 0 1) pl)
  | Ap(Ap(Imp,m1),m2) ->
      let ql = extract_conjuncts m1 true in
      if andlist mizar_type_of_var_p ql then
	mizar_pattern m2 i (pl @ ql)
      else
	raise Not_found
  | Ap(Ap(Eq(rtp),m1),m2) ->
      begin
	let (h,al) = head_spine m1 in
	match h with
	| Name(h1,_) ->
	    let al2 = List.map (fun n -> match n with DB(j,a) when a = Base("$i") -> j | _ -> raise Not_found) al in
	    (h1,i,pl,rtp,al2,m1,m2)
	| _ -> raise Not_found
      end
  | _ -> raise Not_found

	(*** elaboration ***)
let istp = Base("$i")
let oo = Ar(Prop,Prop)
let ooo = Ar(Prop,oo)
let predisj x y = Ap(Ap(Or,x),y)
let preconj x y = Ap(Ap(And,x),y)
let preiff x y = Ap(Ap(Iff,x),y)

let premneg x = MAp(MGround(Neg),x)
let meq a x y = MAp(MAp(MGround(Eq(a)),x),y)

let nxtfreshvar a st =
  let (_,_,sh,_,_,(_,_,varcnt)) = st in
  incr varcnt;
  let en = "#" ^ string_of_int !varcnt in
  let e = Name(en,a) in
  if debug then Printf.printf "#+#++ Created fresh var %s\n" en;
  if a = Base("$i") then Hashtbl.add sh e (Ap(Name("m1_hidden",Ar(Base("$i"),Prop)),e));
  e

let rec abstract_subterms i x m =
  if x = m then
    DB(i,tpof x)
  else
    match m with
    | Ap(m1,m2) -> Ap(abstract_subterms i x m1,abstract_subterms i x m2)
    | Lam(a,m1) -> Lam(a,abstract_subterms (i+1) x m1)
    | _ -> m

exception ElaborationSimpleTypeError of trm * stp * stp
exception ElaborationPatternArityError of string * int * int
exception ElaborationMizarTypeConflict of trm * trm
exception ElaborationFailure of string

let debugwrapper s f x =
  Printf.printf "%s in" s; flush stdout;
  let y = f x in
  Printf.printf "%s out" s; flush stdout;
  y
      
let rec pre_head_spine m ml =
  match m with
  | PAp(m1,m2) -> pre_head_spine m1 (m2::ml)
  | _ -> (m,ml)

let emptydelta : (string,trm) Hashtbl.t = Hashtbl.create 1

let mhead_spine m =
  let rec mhead_spine_r m s =
    match m with
      MAp(m1,m2) -> mhead_spine_r m1 (m2::s)
    | MGround(n) -> (n,s)
    | _ -> raise Not_found
  in
  mhead_spine_r m []

let mneg_body m =
  match m with
  | MAp(MGround(Neg),n) -> Some(m)
  | MAp(MAp(MGround(Imp),n),MGround(False)) -> Some(m)
  | _ -> None

let extract_mizar_type_mprop_atom p =
  match mneg_body p with
  | Some(n) ->
      begin
	let (h,s) = mhead_spine n in
	match s with
        | (a::r) -> (false,a,h,r)
        | _ -> raise Not_found
      end
  | None ->
      begin
	let (h,s) = mhead_spine p in
	match s with
        | (a::r) -> (true,a,h,r)
        | _ -> raise Not_found
      end

let extract_mizar_type_prop_atom p =
  match neg_body p with
  | Some(n) ->
      begin
	let (h,s) = head_spine n in
	match s with
        | (a::r) -> (false,a,h,r)
        | _ -> raise Not_found
      end
  | None ->
      begin
	let (h,s) = head_spine p in
	match s with
        | (a::r) -> (true,a,h,r)
        | _ -> raise Not_found
      end

exception MizarRuleDoesNotApply

    (*** Following Nelson Oppen 1978 Congruence Closure Paper ***)
type cctables = (trm,unit) Hashtbl.t * (trm,trm) Hashtbl.t * (trm,trm) Hashtbl.t * (trm,trm) Hashtbl.t

let print_cctables cc =
  let (ccv,ccp,ccb,ccf) = cc in
  Printf.printf "===== CC ======\n";
  Printf.printf "===== handled ======\n";
  Hashtbl.iter (fun k () -> Printf.printf "%s\n" (trm_str k)) ccv;
  Printf.printf "===== pred ======\n";
  Hashtbl.iter (fun k v -> Printf.printf "%s  ~~>  %s\n" (trm_str k) (trm_str v)) ccp;
  Printf.printf "===== back ======\n";
  Hashtbl.iter (fun k v -> Printf.printf "%s  ~~>  %s\n" (trm_str k) (trm_str v)) ccb;
  Printf.printf "===== find ======\n";
  Hashtbl.iter (fun k v -> Printf.printf "%s  ~~>  %s\n" (trm_str k) (trm_str v)) ccf


let rec mizar_cong_find ccf m =
  try
    let r = mizar_cong_find ccf (Hashtbl.find ccf m) in
    Hashtbl.replace ccf m r;
    r
  with Not_found ->
    m

let mizar_cong_union (ccb,ccf) m n =
  let m1 = mizar_cong_find ccf m in
  let n1 = mizar_cong_find ccf n in
  Hashtbl.add ccf m1 n1;
  Hashtbl.add ccb n1 m1

let mizar_cong_class ccb m =
  let mh : (trm,unit) Hashtbl.t = Hashtbl.create 10 in
  let rec mizar_cong_class_1 m =
    if not (Hashtbl.mem mh m) then
      begin
	Hashtbl.add mh m ();
	List.iter (fun n -> mizar_cong_class_1 n) (Hashtbl.find_all ccb m)
      end
  in
  mizar_cong_class_1 m;
  let ml = ref [] in
  Hashtbl.iter (fun m () -> ml := m::!ml) mh;
  !ml

let mizar_cong_preds (ccp,ccb) m =
  let mh : (trm,unit) Hashtbl.t = Hashtbl.create 10 in
  let ph : (trm,unit) Hashtbl.t = Hashtbl.create 10 in
  let rec mizar_cong_preds_1 m =
    if not (Hashtbl.mem mh m) then
      begin
	List.iter (fun p -> if not (Hashtbl.mem ph p) then Hashtbl.add ph p ()) (Hashtbl.find_all ccp m);
	Hashtbl.add mh m ();
	List.iter (fun n -> mizar_cong_preds_1 n) (Hashtbl.find_all ccb m)
      end
  in
  mizar_cong_preds_1 m;
  let pl = ref [] in
  Hashtbl.iter (fun p () -> pl := p::!pl) ph;
  !pl

let rec mizar_cong_merge (ccp,ccb,ccf) m n =
  if debug then Printf.printf "merging %s with %s\n" (trm_str m) (trm_str n);
  let m1 = mizar_cong_find ccf m in
  let n1 = mizar_cong_find ccf n in
  if m1 = n1 then
    ()
  else
    let pml = mizar_cong_preds (ccp,ccb) m1 in
    let pnl = mizar_cong_preds (ccp,ccb) n1 in
    mizar_cong_union (ccb,ccf) m n;
    List.iter
      (fun p ->
	let p1 = mizar_cong_find ccf p in
	(List.iter
	   (fun q ->
	     let q1 = mizar_cong_find ccf q in
	     if not (p1 = q1) && mizar_cong_congruent (ccp,ccb,ccf) p q then
	       mizar_cong_merge (ccp,ccb,ccf) p q)
	   pnl))
      pml
and mizar_cong_congruent (ccp,ccb,ccf) m n =
  let (mh,ms) = head_spine m in
  let (nh,ns) = head_spine n in
  if mh = nh then
    begin
      try
	if not (List.length ms = List.length ns) then
	  begin
	    Printf.printf "#+#++ mizar_cong_congruent iter2 |ms| = %d ; |ns| = %d ;\nm = %s\nn = %s\n" (List.length ms) (List.length ns) (trm_str m) (trm_str n);
	    raise Exit
	  end
	else
	  List.iter2
	    (fun mi ni ->
	      let mi1 = mizar_cong_find ccf mi in
	      let ni1 = mizar_cong_find ccf ni in
	      if not (mi1 = ni1) then raise Exit)
	    ms ns;
	true
      with Exit -> false
    end
  else
    false

let rec extend_cong_graph (ccv,ccp,ccb,ccf) m =
  if not (Hashtbl.mem ccv m) then
    begin
      if debug then Printf.printf "extend_cong_graph %s\n" (trm_str m);
      let (mh,s) = head_spine m in
      List.iter
	(fun n ->
	  extend_cong_graph (ccv,ccp,ccb,ccf) n;
	  if debug then Printf.printf "back from handling arg %s to extend_cong_graph %s\n" (trm_str n) (trm_str m);
	  let pnl = mizar_cong_preds (ccp,ccb) (mizar_cong_find ccf n) in
	  List.iter
	    (fun p ->
	      if debug then Printf.printf "pred %s in extend_cong_graph %s\n" (trm_str p) (trm_str m);
	      let p1 = mizar_cong_find ccf p in
	      if debug then Printf.printf "pred find %s\n" (trm_str p1);
	      let m1 = mizar_cong_find ccf m in
	      if debug then Printf.printf "m find %s\n" (trm_str m1);
	      if not (p1 = m1) && mizar_cong_congruent (ccp,ccb,ccf) p m then
		mizar_cong_merge (ccp,ccb,ccf) p m)
	    pnl;
	  Hashtbl.add ccp n m)
	s;
      Hashtbl.add ccv m ()
    end

let mizar_cong cc m1 m2 =
  extend_cong_graph cc m1;
  extend_cong_graph cc m2;
  let (_,_,_,ccf) = cc in
  let n1 = mizar_cong_find ccf m1 in
  let n2 = mizar_cong_find ccf m2 in
  if debug then
    begin
      if not (m1 = n1) then Printf.printf "cong_find\n   %s\n ~> %s\n" (trm_str m1) (trm_str n1);
      if not (m2 = n2) then Printf.printf "cong_find\n   %s\n ~> %s\n" (trm_str m2) (trm_str n2);
    end;
  n1 = n2
	
type mizstate = (pretrm,trm * stp * ((string * (trm list)) list)) Hashtbl.t * (trm,trm) Hashtbl.t * (trm,trm) Hashtbl.t * (trm,unit) Hashtbl.t * cctables * ((string,unit) Hashtbl.t * bool ref * int ref)
    
let common_mizar_type_p p =
  let (pol,a,m,r) = extract_mizar_type_prop_atom p in  
  if pol then (*** positive ***)
    begin
      match m with (*** could be more specific ***)
      | Name(x,_) -> List.mem x ["v13_vectsp_1";"v2_card_3";"m1_group_2";"v2_yellow21";"l1_cgames_1";"v3_ordinal1";"v8_bcialg_1";"v1_afvect0";"v4_unialg_1";"v2_pre_topc";"v11_altcat_1";"m1_group_9";"v2_collsp";"v12_altcat_1";"m1_finseq_1";"m1_rusub_1";"m1_rlsub_1";"v2_matrixj1";"v1_polynom1";"v24_waybel_0";"m1_pre_topc";"v2_msualg_4";"v3_cat_5";"v3_bcialg_1";"v3_orders_2";"v5_orders_2";"v1_borsuk_2";"v4_circcmb3";"v1_xreal_0";"m1_rmod_2";"v4_relat_1";"l1_altcat_1";"m1_pythtrip";"v12_osalg_1";"v1_msualg_4";"v2_pre_poly";"v13_algstr_0";"l1_metric_1";"m1_zmodul01";"v1_glib_003";"v5_ordinal1";"v1_finseq_1";"v2_membered";"v1_funct_2";"v1_xxreal_0";"v6_membered";"m1_vectsp_4";"l1_bhsp_1";"l1_rlvect_1";"v2_group_1";"v16_pcs_0";"v2_finseq_1";"l1_orders_2";"v1_int_1";"v1_quaterni";"v2_pralg_1";"v1_ordinal2";"v1_partfun1";"v3_card_1";"v1_trees_1";"l1_graph_1";"v2_compts_1";"v1_group_3";"l3_algstr_0";"m1_subset_1";"v4_valued_0";"v1_xcmplx_0";"v2_memstr_0";"v7_pcs_0";"l1_struct_0";"l1_grcat_1";"l2_algstr_0";"l1_zmodul01";"l1_pre_topc";"l1_unialg_1";"v5_finset_1";"v8_osalg_1";"v1_membered";"v1_bilinear";"v1_finset_1";"l1_ringcat1";"v2_margrel1";"l6_algstr_0";"v1_relat_1";"m1_hidden";"v1_funct_1";"v7_ordinal1";"v3_finseq_1"]
      | _ -> false
    end
  else (*** negative ***)
    begin
      match m with (*** could be more specific ***)
      | Name(x,_) -> List.mem x ["v1_xboole_0";"v2_struct_0";"v3_funct_1";"v3_pencil_1";"v7_struct_0";"v3_xxreal_0";"v6_struct_0";"v1_setfam_1"]
      | _ -> false
    end

let rec append_consis ml nl =
  match ml with
  | m::mr ->
      let nm = normneg m in
      if List.mem nm nl then
	raise (ElaborationFailure(Printf.sprintf "assigned + and - sort %s" (trm_str m)))
      else
	append_consis mr (m::nl)
  | [] -> nl

let rec adjoin_frees x ml frl =
  match frl with
  | [] -> [(x,ml)]
  | ((y,nl)::frr) when x = y -> ((x,append_consis ml nl)::frr)
  | ((y,nl)::frr) -> (y,nl)::adjoin_frees x ml frr

let rec union_frees frl1 frl2 =
  match frl1 with
  | [] -> frl2
  | ((x,ml)::frr1) -> union_frees frr1 (adjoin_frees x ml frl2)

let replsep_p m =
  let (h,s) = head_spine m in
  match h with
  | Name(z,_) when List.mem z ["replSep1";"replSep2";"replSep3";"replSep4";"replSep5";"replSep6"] -> true
  | _ -> false
  
let rec known_mizar_type_fact st p =
  let (_,_,sh,_,cc,_) = st in
  try
    let (pol,a,m,r) = extract_mizar_type_prop_atom p in (*** if it's marked as a variable, then just assign it the type ***)
    if replsep_p a then
      pol && (match m with Name(m1h,_) when m1h = "m1_hidden" -> true | _ -> false) && r = []
    else
      begin (*** ask about all equivalent forms of a ***)
	let (ccv,ccp,ccb,ccf) = cc in
	if debug then Printf.printf "Extending cong graph with p = %s\n" (trm_str p);
	extend_cong_graph cc p;
	extend_cong_graph cc a;
	let p1 = mizar_cong_find ccf p in
	if debug then Printf.printf "p1 = %s\n" (trm_str p1);
	let a1 = mizar_cong_find ccf a in
	let al = mizar_cong_class ccb a1 in
	List.iter (fun a -> all_mizar_sorts st a) al;
	if debug then
	  begin
	    if p1 = p then
	      Printf.printf "Asking if %s is a known mizar type fact\nEquivalent forms:\n" (trm_str p)
	    else
	      Printf.printf "Asking if %s (in equiv form %s) is a known mizar type fact\nEquivalent forms:\n" (trm_str p) (trm_str p1);
	    List.iter (fun a -> Printf.printf "~ %s\n" (trm_str a)) al;
	  end;
	try
	  List.iter
	    (fun a ->
	      List.iter
		(fun af ->
		  if debug then Printf.printf "Extending cong graph with af = %s\n" (trm_str af);
		  extend_cong_graph cc af;
		  let af1 = mizar_cong_find ccf af in
		  if debug then (if af = af1 then Printf.printf ". %s\n" (trm_str af) else Printf.printf ". %s\n~ %s\n" (trm_str af) (trm_str af1));
		  let p2 = mizar_cong_find ccf p1 in (*** since we've extended the cong graph, p1's NF may have changed ***)
		  if debug then (Printf.printf "Comparing af1 %s\nwith p2 %s\n" (trm_str af1) (trm_str p2));
		  if af1 = p2 then raise Exit)
		(Hashtbl.find_all sh a))
	    al;
          false
	with Exit -> true
      end
  with Not_found -> false (*** this shouldn't happen, but if it does just say it's not a fact ***)
and known_mizar_type_fact_assumeifunknown st p frl =
  let (_,_,sh,_,cc,_) = st in
  try
    let (pol,a,m,r) = extract_mizar_type_prop_atom p in (*** if it's marked as a variable, then just assign it the type ***)
    begin (*** ask about all equivalent forms of a ***)
      let (ccv,ccp,ccb,ccf) = cc in
      extend_cong_graph cc p;
      extend_cong_graph cc a;
      let p1 = mizar_cong_find ccf p in
      let a1 = mizar_cong_find ccf a in
      let al = mizar_cong_class ccb a1 in
      List.iter (fun a -> all_mizar_sorts st a) al;
      if debug then
	begin
	  if p1 = p then
	    Printf.printf "Asking if %s is a known mizar type fact\nEquivalent forms of a:\n" (trm_str p)
	  else
	    Printf.printf "Asking if %s (in equiv form %s) is a known mizar type fact\nEquivalent forms of a:\n" (trm_str p) (trm_str p1);
	  List.iter (fun a -> Printf.printf "~ %s\n" (trm_str a)) al;
	end;
      try
	List.iter
	  (fun a ->
	    List.iter
	      (fun af ->
		extend_cong_graph cc af;
		let af1 = mizar_cong_find ccf af in
		if debug then (if af = af1 then Printf.printf ". %s\n" (trm_str af) else Printf.printf ". %s\n~ %s\n" (trm_str af) (trm_str af1));
		if af1 = p1 then raise Exit)
	      (Hashtbl.find_all sh a))
	  al;
	match a with
	| Name(x,Base("$i")) ->
	    begin
	      try
		let ml = List.assoc x frl in
		if !raiseunknownname then raise (UnknownName(x));
		if debug then Printf.printf "%s is unknown, seeing if type can be assigned\n" (trm_str a);
		let np = normneg p in
		if List.mem np ml then
		  None
		else
		  Some(adjoin_frees x [p] frl)
	      with Not_found -> None
	    end
	| _ -> None
      with Exit -> Some(frl)
    end
  with Not_found -> None (*** this shouldn't happen, but if it does just say it's not a fact ***)
and mizar_adjust st p lim progress = (*** apply redefinitions when possible ***)
  if lim <= 0 then (*** bound ***)
    ()
  else
    mizar_adjusted_facts st p progress
and mizar_roundup st p pol a m r lim progress = (*** apply clusters when possible ***)
  if lim <= 0 then (*** bound ***)
    ()
  else
    match m with
    | Name(mx,_) ->
	List.iter
	  (fun (ax,i,prim,pl,bl) ->
	    if debug then
	      begin
		Printf.printf "Trying generic forward typing rule %s, p is %s\n" ax (trm_str p);
		Printf.printf "Primary prem: %s\n" (trm_str prim);
		List.iter (fun p2 -> Printf.printf "Secondary prem: %s\n" (trm_str p2)) pl;
	      end;
	    if mizar_apply_typing_rule ax st i pl bl prim p lim then progress := true)
	  (Hashtbl.find_all mizar_typing_gen_forward mx);
	begin
	  let (h,s) = head_spine a in
	  match h with
	  | Name(hx,_) ->
	      List.iter
		(fun (ax,i,gena,pl,concl) ->
		  if debug then
		    begin
		      Printf.printf "Trying generic backward typing rule %s, p is %s\n" ax (trm_str p);
		      Printf.printf "Primary term: %s\n" (trm_str a);
		      Printf.printf "Concl: %s\n" (trm_str concl);
		      List.iter (fun p2 -> Printf.printf "Prem: %s\n" (trm_str p2)) pl;
		    end;
		  try
		    let tau = mizar_rule_applies st ax i pl gena a (Base("$i")) in
		    if debug then Printf.printf "%s applies; applying it now\n" ax;
		    mizar_rule_apply st [concl] tau lim;
		    progress := true
		  with MizarRuleDoesNotApply -> ())
		(Hashtbl.find_all mizar_typing_gen_backward hx)
	  | _ -> ()
	end
    | _ -> ()
and mizar_widen st pl lim =
  if lim <= 0 then (*** bound ***)
    ()
  else
    try
      if debug then Printf.printf "====== mizar_widen [limit %d]\n" lim;
      let (_,_,sh,_,cc,_) = st in
      let (ccv,ccp,ccb,ccf) = cc in
      let polamrl = List.map (fun p -> let (pol,a,m,r) = extract_mizar_type_prop_atom p in (p,pol,a,m,r)) pl in
      let polamrnewl = List.filter (fun (p,pol,a,m,r) -> not (List.mem p (Hashtbl.find_all sh a))) polamrl in
      List.iter
	(fun (p,pol,a,m,r) ->
	  if debug then Printf.printf "New Computed Mizar Sort for %s: %s\n" (trm_str a) (trm_str p);
	  let np = normneg p in
	  if List.mem np (Hashtbl.find_all sh a) then
	    begin
	      if debug then Printf.printf "Contradicts Previous Mizar Sort for %s: %s\n" (trm_str a) (trm_str np);
	      raise (ElaborationFailure(Printf.sprintf "%s assigned + and - sort %s" (trm_str a) (trm_str p)))
	    end;
	  Hashtbl.add sh a p;
	  let m1hidden = Ap(Name("m1_hidden",Ar(Base("$i"),Prop)),a) in
	  if not (List.mem m1hidden (Hashtbl.find_all sh a)) then Hashtbl.add sh a m1hidden) (*** hack: once a has some sort, also give it sort m1_hidden (set); look at dt_k3_topgen_3 to see why this is needed ***)
	polamrnewl;
      let repeatcnt = ref 0 in
      let progress = ref true in
      while !repeatcnt < !maxelaborateiters && !progress do (*** repeat the computation of new types several times as long as new info has been found, up to a limit ***)
	if debug then Printf.printf "Iteration %d of computation of new sorts:\n" !repeatcnt;
	incr repeatcnt;
	progress := false;
	List.iter
	  (fun (p,pol,a,m,r) ->
	    begin (*** climb the hierarchy ***)
	      match m with
	      | Name(mx,_) ->
		  List.iter
		    (fun (ax,i,prim,pl,bl) ->
		      if debug then
			begin
			  Printf.printf "Climbing hierarchy using %s\n" ax;
			  Printf.printf "Primary prem: %s\n" (trm_str prim);
			  List.iter (fun p2 -> Printf.printf "Secondary prem: %s\n" (trm_str p2)) pl;
			end;
		      if mizar_apply_typing_rule ax st i pl bl prim p lim then progress := true)
		    (Hashtbl.find_all mizar_typing mx)
	      | _ -> ()
	    end;
	    if debug then Printf.printf "====== hierarchy climbed %s [limit %d]\n" (trm_str p) lim;
	    mizar_adjust st p (lim-1) progress;
	    if debug then Printf.printf "====== after adjust %s [limit %d]\n" (trm_str p) lim;
	    if debug then Printf.printf "====== now rounding up %s [limit %d]\n" (trm_str p) lim;
	    mizar_roundup st p pol a m r (lim-1) progress)
	  polamrnewl;
	let congforms = ref [] in
	List.iter
	  (fun p ->
	    if debug then Printf.printf "Looking for new congruent forms of %s\n" (trm_str p);
	    extend_cong_graph cc p;
	    let p1 = mizar_cong_find ccf p in
	    let pl2 = mizar_cong_class ccb p1 in
	    List.iter
	      (fun p2 ->
		if debug then Printf.printf "Computed congruent %s\n" (trm_str p2);
		if not (List.mem p2 pl) then congforms := p2::!congforms)
	      pl2)
	  pl;
	if not (!congforms = []) then
	  begin
	    if debug then Printf.printf "%d new congruent forms of typing facts computed; widening with those\n" (List.length !congforms);
	    mizar_widen st !congforms (lim-1)
	  end
      done
    with Not_found -> () (*** should not happen ***)
and mizar_rule_applies st ax i pl gen spec a =
  let tau = ref [] in
  for j = 0 to i-1 do
    let (_,n) = new_evar 0 [] istp in
    tau := n::!tau
  done;
  try
    let genx = metanorm (meta_simulsubst_meta (to_meta gen) !tau) in
    if debug then Printf.printf "About to match\n%s\nagainst\n%s\n" (metatrm_str genx) (trm_str spec);
    let (_,_,sh,_,cc,_) = st in
    let (ccv,ccp,ccb,ccf) = cc in
    let dpairs = pattern_match (fun m -> extend_cong_graph cc m; mizar_cong_class ccb (mizar_cong_find ccf m)) [([],genx,spec,a)] in
    if dpairs = [] then (*** nothing should be leftover ***)
      begin
	List.iter
	  (fun p1 ->
	    let p2 = metanorm (meta_simulsubst_meta (to_meta p1) !tau) in
	    try
	      let p3 = meta_to_ground emptydelta p2 in
	      if not (known_mizar_type_fact st p3) then
		begin
		  if debug then Printf.printf "Rule not applied since premise not known: %s\n" (trm_str p3);
		  if debug then Printf.printf "(NOTAPPLYBECAUSE \"%s\" \"%s\")\n" ax (trm_str p3);
		  raise MizarRuleDoesNotApply
		end
	    with NotGround ->
	      if debug then Printf.printf "Rule not applied since premise not ground: %s\nThis probably shouldn't be a rule, since the primary lit apparently didn't contain all the vars.\n" (metatrm_str p2);
	      raise MizarRuleDoesNotApply)
	  pl;
	    (*** if here, then all premises were known, so the rule applies ***)
	!tau
      end
    else
      begin
	if debug then Printf.printf "leftover dpairs; not applying rule\n";
	raise MizarRuleDoesNotApply
      end
  with PatternMatchFailed ->
    if debug then Printf.printf "pattern match failed; not applying rule\n";
    raise MizarRuleDoesNotApply
and mizar_add_to_congclos st m1 m2 = (*** to do, update fourth component of st ***)
  let (_,_,_,_,cc,_) = st in
  let (ccv,ccp,ccb,ccf) = cc in
  extend_cong_graph cc m1;
  extend_cong_graph cc m2;
  mizar_cong_merge (ccp,ccb,ccf) m1 m2
and mizar_adjusted_terms st m progress =
  let (h,s) = head_spine m in
  begin
    List.iter (fun m -> mizar_adjusted_terms st m progress) s;
    match h with
    | Name(x,_) ->
	begin
	  try
	    let (ax,i,pl,al,d) = Hashtbl.find mizar_redef x in
	    let gen = ref h in
	    List.iter (fun a -> gen := Ap(!gen,a)) al;
	    if debug then Printf.printf "Trying to apply term redef %s\n" ax;
	    let tau = mizar_rule_applies st ax i pl !gen m (Base("$i")) in
	    if debug then Printf.printf "redef %s applies\n" ax;
	    let d2 = metanorm (meta_simulsubst_meta (to_meta d) tau) in
	    let d3 = meta_to_ground emptydelta d2 in
	    if debug then Printf.printf "redef gives %s == %s\n" (trm_str m) (trm_str d3);
	    if not (m = d3) then all_mizar_sorts st d3;
	    mizar_add_to_congclos st m d3;
	    progress := true
	  with
	  | Not_found -> ()
	  | MizarRuleDoesNotApply -> ()
	  | NotGround -> ()
	end
    | _ -> ()
  end
and mizar_adjusted_atoms st a progress =
  if debug then Printf.printf "mizar_adjusted_atoms %s\n" (trm_str a);
  let (h,s) = head_spine a in
  begin
    if debug then Printf.printf "mizar_adjusted_atoms h = %s\n" (trm_str h);
    List.iter (fun m -> mizar_adjusted_terms st m progress) s;
    match h with
    | Name(x,_) ->
	begin
	  try
	    let (ax,i,pl,al,d) = Hashtbl.find mizar_redef x in
	    if debug then Printf.printf "Trying to apply reln redef %s\n" ax;
	    let gen = ref h in
	    List.iter (fun a -> gen := Ap(!gen,a)) al;
	    let tau = mizar_rule_applies st ax i pl !gen a Prop in
	    if debug then Printf.printf "redef %s applies\n" ax;
	    let d2 = metanorm (meta_simulsubst_meta (to_meta d) tau) in
	    let d3 = meta_to_ground emptydelta d2 in
	    if debug then Printf.printf "redef gives %s == %s\n" (trm_str a) (trm_str d3);
	    mizar_add_to_congclos st a d3;
	    progress := true
	  with
	  | Not_found -> ()
	  | MizarRuleDoesNotApply -> ()
	  | NotGround -> ()	    
	end
    | _ -> ()
  end
and mizar_adjusted_facts st p progress =
  if debug then Printf.printf "mizar_adjusted_facts %s\n" (trm_str p);
  match neg_body p with
  | Some(n) ->
      mizar_adjusted_atoms st n progress
  | None ->
      mizar_adjusted_atoms st p progress
and mizar_rule_apply st bl tau lim =
  let towiden = ref [] in
  List.iter
    (fun b ->
      let b2 = metanorm (meta_simulsubst_meta (to_meta b) tau) in
      try
	let b3 = meta_to_ground emptydelta b2 in
	towiden := b3::!towiden
      with NotGround ->
	if debug then Printf.printf "Conclusion of rule was not ground: %s\nThis probably shouldn't be a rule, since the primary lit apparently didn't contain all the vars.\n" (metatrm_str b2);
	raise Exit)
    bl;
  mizar_widen st !towiden lim
and mizar_apply_typing_rule ax st i pl bl prim p lim = (*** in this function assume there are no missing args -- it's for typing not elaboration ***)
  if lim <= 0 then
    (if debug then Printf.printf "limit reached, not applying %s\n" ax; false)
  else
    begin
      try
	if debug then Printf.printf "Testing if %s applies\n" ax;
	let tau = mizar_rule_applies st ax i pl prim p Prop in
	if debug then Printf.printf "%s applies; applying it\n" ax;
	    (*** if here, then all premises were known, so the rule applies ***)
	mizar_rule_apply st bl tau lim;
	true
      with
      | Exit ->
	  if debug then Printf.printf "Exit? %s does not apply\n" ax;
	  false
      | MizarRuleDoesNotApply ->
	  if debug then Printf.printf "%s does not apply\n" ax;
	  false
    end
and func_mizar_sorts st h s =
  match h with
  | Name(f,_) ->
      begin
	let ls = List.length s in
	let towiden = ref [] in
	let ftl = Hashtbl.find_all mizar_func_types f in
	if debug then Printf.printf "trying to find types for %s (|ftl|=%d)\n" f (List.length ftl);
	List.iter
	  (fun (ax,n,pl,ml) ->
	    if not (n = ls) then
	      (Printf.printf "*** arity mismatch while trying to use %s (%d) for %s, rethink code for func_mizar_sorts\n" ax n f)
	    else
	      begin
		if debug then Printf.printf "trying to use %s (%d) for %s\n" ax n f;
		try
		  ignore (List.find
			    (fun p ->
			      let p1 = simulsubst p s in
			      let b1 = known_mizar_type_fact st p1 in
			      not b1)
			    pl);
	      (*** if an unmatched premise was found, don't apply the rule ***)
		  ()
		with Not_found ->
	      (*** all premises were matched, apply the rule ***)
		  if debug then Printf.printf "applying %s for %s\n" ax f;
		  List.iter
		    (fun m1 ->
		      let m2 = simulsubst m1 s in
		      towiden := m2::!towiden)
		    ml
	      end)
	  ftl;
	mizar_widen st !towiden !elaboratelim
      end
  | _ -> ()
and all_mizar_sorts st m =
  let (_,_,sh,sc,_,_) = st in
  if not (Hashtbl.mem sc m) then 
    begin
      if debug then Printf.printf "About to compute all Mizar sorts for %s\n" (trm_str m);
      Hashtbl.add sc m ();
      let (h,s) = head_spine m in
      let rs = List.rev s in
      List.iter
	(fun n -> all_mizar_sorts st n)
	s;
      func_mizar_sorts st h rs;
      if debug then Printf.printf "Finished computing all Mizar sorts for %s\n" (trm_str m);
    end
	
let assume_mizar_sorts st g m =
  let ql = extract_conjuncts m true in
  let towiden = ref [] in
  List.iter
    (fun q ->
      try
	let (pol,a,mh,r) = extract_mizar_type_prop_atom q in
	ignore (List.find (fun (x,xx) -> a = xx) g);
	if debug then Printf.printf "assuming sorting fact for %s: %s\n" (trm_str a) (trm_str m);
	towiden := m::!towiden
      with Not_found -> ())
    ql;
  mizar_widen st !towiden !elaboratelim

let alldigits_p n =
  try
    for i = 0 to (String.length n) - 1 do
      let c = Char.code n.[i] in
      if (c < 48) || (c > 57) then raise Exit
    done;
    true
  with Exit -> false
      
let num_p n =
  String.length n > 0 && alldigits_p n

let unknown_name_p st g x =
  not (num_p x)
    &&
  not (Hashtbl.mem !mizar_name_tp x)
    &&
  not (try ignore (List.find (fun (_,n) -> match n with Name(xx,_) when xx = x -> true | _ -> false) g); true with Not_found -> false)

let fake_var_sort_p p g =
  match neg_body p with
  | Some(q) ->
      begin
	match head_spine q with
	| (_,(y::_)) ->
	    begin
	      try
		let (x,_) = List.find (fun (x,z) -> z = y) g in
		x = ""
	      with Not_found ->
		false
	    end
	| _ -> false
      end
  | None ->
      begin
	match head_spine p with
	| (_,(y::_)) ->
	    begin
	      try
		let (x,_) = List.find (fun (x,z) -> z = y) g in
		x = ""
	      with Not_found ->
		false
	    end
	| _ -> false
      end

let rec mizar_elaborate_spine st g h (ml:pretrm list) frl : trm * stp * ((string * (trm list)) list) =
  if debug then Printf.printf "mizar_elaborate_spine %s %d\n" (pretrm_str h) (List.length ml);
  if debug then
    begin
      Printf.printf "#+#++ frl:\n";
      List.iter
	(fun (x,ml) ->
	  Printf.printf "#+#++ %s :\n" x;
	  List.iter (fun m -> Printf.printf " %s;\n" (trm_str m)) ml;
	  Printf.printf "\n")
	frl;
    end;
  let (_,eh,sh,sc,cc,_) = st in
  let (ccv,ccp,ccb,ccf) = cc in
  match h with
   | PName x ->
      begin
	try
	  let (mh,mhtp) = mizar_elaborate_trm_var x g in
	  mizar_elaborate_spine_2 st g mh mhtp ml frl
	with Not_found ->
	  try
	    let (i,pl,rtp,al,pd,d) = Hashtbl.find mizar_pattern_table x in
	    if debug then Printf.printf "found %s as pattern with %d args, %d visible\nd = %s\n" x i (List.length al) (trm_str d);
	    if List.length al = List.length ml then
	      begin
		let tau = ref [] in
		for j = 0 to i-1 do
		  let (_,n) = new_evar 0 [] istp in
		  tau := n::!tau
		done;
		let (_,explargs) = head_spine pd in
		let dpairs = ref [] in
		let frlr = ref frl in
		begin (** keep looping as long as progress; progress if at least one sort premiss is matched; fail if a ground premise is not known already; fail if no progress; continue if all ground **)
		  try
		    if not (List.length explargs = List.length ml) then
		      begin
			Printf.printf "#+#++ mizar_elaborate_spine iter2 |explargs| = %d ; |ml| = %d\n" (List.length explargs) (List.length ml);
		      end;
		    List.iter2
		      (fun explarg m ->
			let (n,frl) = mizar_elaborate_trm_force_stp st g m istp in
			frlr := union_frees !frlr frl;
			dpairs := pattern_match (fun m -> extend_cong_graph cc m; mizar_cong_class ccb (mizar_cong_find ccf m)) (([],metanorm (meta_simulsubst_meta (to_meta explarg) !tau),n,istp)::!dpairs))
		      explargs
		      ml;
		  with PatternMatchFailed ->
		    raise (ElaborationFailure("something went wrong when trying to elaborate missing args of " ^ x))
		end;
                if debug then Printf.printf "done with visible args, and back to pattern %s\n" x;
		if debug then
		  begin
		    Printf.printf "#+#++ frlr:\n";
		    List.iter
		      (fun (x,ml) ->
			Printf.printf "#+#++ %s :\n" x;
			List.iter (fun m -> Printf.printf " %s;\n" (trm_str m)) ml;
			Printf.printf "\n")
		      !frlr;
		  end;
		let progress = ref true in
		while !progress do
		  progress := false;
		  List.iter
		    (fun p ->
		      if debug then Printf.printf "%s progress loop 1 p = %s\n" x (trm_str p);
		      let p2 = metanorm (meta_simulsubst_meta (to_meta p) !tau) in
		      try
			if debug then Printf.printf "progress loop 1b p2 = %s\n" (metatrm_str p2);
			let p3 = meta_to_ground emptydelta p2 in
			if debug then Printf.printf "progress loop 1c p3 = %s\n" (trm_str p3);
			let (_,_,_,_,_,(_,u,_)) = st in
			if !u || not (!acceptcommonmizartypes && common_mizar_type_p p) then (*** if acceptcommonmizartypes is set, then don't bother checking or enforcing these preconditions on mizar patterns ***)
			  begin
			    if not (Hashtbl.mem main_queried_sorts p3) then Hashtbl.add main_queried_sorts p3 ();
			    if not (known_mizar_type_fact st p3) && not (Hashtbl.mem main_assumed_sorts p3) then
			      begin
				if debug then Printf.printf "*1** %d Assuming sort %s\n" !assume_needed_sorts (trm_str p3);
				if !assume_needed_sorts > 0 then raise NeedToAssume;
				if not (fake_var_sort_p p3 g) then Hashtbl.add main_assumed_sorts p3 ();
			      end;
			  end
		      with NotGround ->
			if debug then
			  begin
			    Printf.printf "%s Trying to match pattern premise sort %s\n" x (metatrm_str p2);
			    Printf.printf "p = %s\n" (trm_str p);
			    Printf.printf "tau:\n";
			    List.iter (fun tau1 -> Printf.printf " %s\n" (metatrm_str tau1)) !tau;
			  end;
			try
			  let (pos,n,h,r) = extract_mizar_type_mprop_atom p2 in
			  let n2 = meta_to_ground emptydelta n in
			  List.iter
			    (fun a ->
			      try
				if debug then Printf.printf "Trying to match against %s\n" (trm_str a);
				let dpairs2 = pattern_match (fun m -> extend_cong_graph cc m; mizar_cong_class ccb (mizar_cong_find ccf m)) (([],p2,a,Prop)::!dpairs) in
				if debug then Printf.printf "succeeded\n";
				dpairs := dpairs2;
				progress := true;
				raise Exit (*** found a way to say p is true, possibly filling in some missing args; if this is the wrong way to fill in some missing args it may cause elaboration failure ***)
			      with PatternMatchFailed -> ())
			    (Hashtbl.find_all sh n2)
			with
			| NotGround -> ()
			| Not_found -> ()
			| Exit -> ())
		    pl;
		  List.iter
		    (fun n ->
		      try
			if debug then Printf.printf "%s progress loop 2 n = %s\n" x (metatrm_str n);
			let n2 = meta_to_ground emptydelta n in
			if debug then Printf.printf "%s progress loop 2b n2 = %s\n" x (trm_str n2);
			all_mizar_sorts st n2
		      with NotGround -> ())
		    !tau
		done;
		if debug then Printf.printf "Hopefully all missing arguments have been filled in by now (%s)\n" x;
		(*** hopefully all missing arguments have been filled in by now; otherwise fail ***)
		try
		  let tau2 =
		    List.map
		      (fun n ->
			let n2 = meta_to_ground emptydelta n in
                        all_mizar_sorts st n2;
			n2)
		      !tau
		  in (*** Now check how many of the premise sorts are satisfied ***)
		  List.iter
		    (fun p ->
		      begin
			let p2 = simulsubst p tau2 in
			if debug then Printf.printf "%s Trying to match ground pattern premise sort %s\n" x (trm_str p2);
			if not (Hashtbl.mem main_queried_sorts p2) then Hashtbl.add main_queried_sorts p2 ();
			if not (known_mizar_type_fact st p2) && not (Hashtbl.mem main_assumed_sorts p2) then
			  begin
			    if debug then Printf.printf "*2** Assuming sort %s\n" (trm_str p2);
			    if !assume_needed_sorts > 0 then raise NeedToAssume;
			    if not (fake_var_sort_p p2 g) then Hashtbl.add main_assumed_sorts p2 ();
			  end;
		      end)
		    pl;
		  let d2 = simulsubst d tau2 in
		  all_mizar_sorts st d2;
		  (d2,rtp,!frlr)
		with NotGround ->
		  raise (ElaborationFailure("could not fill in all missing arguments of " ^ (metatrm_str (meta_simulsubst_meta (to_meta d) !tau))))
	      end
	    else
	      begin
(*		List.iter (fun m -> Printf.printf ":: %s\n" (pretrm_str m)) ml; *)
		raise (ElaborationPatternArityError(x,i,List.length ml))
	      end
	  with Not_found -> (*** treat 'new' names as variables of type $i, sort m1_hidden, if not applied, otherwise fail ***)
            try
	      if num_p x then
		mizar_elaborate_spine_2 st g (Name(x,Base("$i"))) (Base("$i")) ml frl
	      else
		let a = Hashtbl.find !mizar_name_tp x in
		mizar_elaborate_spine_2 st g (Name(x,a)) a ml frl
            with Not_found ->
              if ml = [] then
	        let n = Name(x,istp) in
		begin
		  if List.mem_assoc x frl then
		    (n,istp,frl)
		  else
		    begin
		      if !raiseunknownname then raise (UnknownName(x));
		      let (_,_,_,_,_,(_,u,_)) = st in
		      if debug then Printf.printf "#+#++ %s is being treated as a free var (1)\n" x;
		      u := true;
	              (n,istp,((x,[])::frl))
		    end
		end
  	      else
                raise (ElaborationFailure("Unknown applied name " ^ x))
   end
  | _ ->
      let (mh,mhtp,frl) = mizar_elaborate_trm st g h in
      mizar_elaborate_spine_2 st g mh mhtp ml frl
and mizar_elaborate_spine_2 st g h tp ml frl : trm * stp * ((string * (trm list)) list) =
  if debug then Printf.printf "mizar_elaborate_spine_2 %s %d\n" (trm_str h) (List.length ml);
  match (tp,ml) with
  | (_,[]) -> func_mizar_sorts st h []; (h,tp,frl)
  | (Ar(a,b),(m::mr)) ->
      if debug then Printf.printf "mizar_elaborate_spine_2 cons %s %d\n" (pretrm_str m) (List.length mr);
      let (n,frl) = mizar_elaborate_trm_force_stp st g m a in
      mizar_elaborate_spine_2 st g (Ap(h,n)) b mr frl
  | _ -> raise (ElaborationSimpleTypeError(h,tp,Ar(Base("*"),Base("*"))))
and mizar_elaborate_spine_other st g h tp ml frl : trm * stp * ((string * (trm list)) list) =
  if debug then Printf.printf "mizar_elaborate_spine_other %s %d\n" (trm_str h) (List.length ml);
  match (tp,ml) with
  | (_,[]) -> (h,tp,frl)
  | (Ar(a,b),(m::mr)) ->
      let (n,frl2) = mizar_elaborate_trm_force_stp st g m a in
      mizar_elaborate_spine_other st g (Ap(h,n)) b mr (union_frees frl frl2)
  | _ -> raise (ElaborationSimpleTypeError(h,tp,Ar(Base("*"),Base("*"))))
and mizar_elaborate_trm (st: mizstate) (g:(string * trm) list) (m:pretrm) : trm * stp * ((string * (trm list)) list) =
  if debug then
    begin
      Printf.printf "mizar_elaborate_trm %s\n" (pretrm_str m);
      Printf.printf "g:\n";
      List.iter (fun (y,n) -> Printf.printf "%s ~> %s\n" y (trm_str n)) g
    end;
  let (ta,_,_,_,_,_) = st in
  try
    let (n,a,frl) = Hashtbl.find ta m in
    (n,a,frl)
  with Not_found ->
    let (n,a,frl) = mizar_elaborate_trm_a st g m in
    Hashtbl.add ta m (n,a,frl);
    (n,a,frl)
and mizar_elaborate_trm_a (st: mizstate) (g:(string * trm) list) (m:pretrm) : trm * stp * ((string * (trm list)) list) =
  match m with
  | PTrm(n) ->
      let frl = ref [] in
      List.iter
	(fun (x,a) ->
	  if unknown_name_p st g x then
	    begin
	      if !raiseunknownname then raise (UnknownName(x));
	      if not (a = Base("$i")) then raise (ElaborationFailure("Unknown name " ^ x));
	      let (_,_,_,_,_,(_,u,_)) = st in
	      if debug then Printf.printf "#+#++ %s is being treated as a free var (2)\n" x;
	      u := true;
	      frl := (x,[])::!frl
	    end)
	(consts_of_trm [] n);
      (n,tpof n,!frl)
  | PName x ->
      begin
	try
	  if num_p x then
	    (Name(x,Base("$i")),Base("$i"),[])
	  else
	    let (n,a) = mizar_elaborate_trm_var x g in
	    (n,a,[])
	with Not_found ->
	  try
	    let (i,pl,rtp,al,pd,d) = Hashtbl.find mizar_pattern_table x in
	    if debug then Printf.printf "found %s as pattern with %d args\n" x i;
	    if i > 0 then (*** in case there is no spine, arity of the n* should be zero ***)
	      raise (ElaborationPatternArityError(x,i,0))
	    else
	      begin
		all_mizar_sorts st d;
		(d,rtp,[])
	      end
	  with Not_found -> (*** treat 'new' names as variables of type $i ***)
	    let n = Name(x,istp) in
	    if !raiseunknownname then raise (UnknownName(x));
	    let (_,_,_,_,_,(_,u,_)) = st in
	    if debug then Printf.printf "#+#++ %s is being treated as a free var (3)\n" x;
	    u := true;
	    (n,istp,[(x,[])])
      end
  | PTrue -> (True,Prop,[])
  | PFalse -> (False,Prop,[])
  | PNeg -> (Neg,oo,[])
  | POr -> (Or,ooo,[])
  | PAnd -> (And,ooo,[])
  | PIff -> (Iff,ooo,[])
  | PAp(PExists,PLam(xl,m1)) -> let (n,frl) = mizar_elaborate_trm_exists st g xl m1 in (n,Prop,frl)
  | PAp(PAp(PName "sort",m1),m2) ->
      let (h,ml) = pre_head_spine m2 [] in
      mizar_elaborate_spine st g h (m1::ml) []
   | PAp(PAp(PImplies,m1),m2) ->
      let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 Prop in
      let n1l = extract_conjuncts n1 true in
      if debug then List.iter (fun n -> Printf.printf "n1l[]: %s\n" (trm_str n); assume_mizar_sorts st g n) n1l;
      let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 Prop in
      (Ap(Ap(Imp,n1),n2),Prop,union_frees frl1 frl2)
   | PAp(PAp(PAnd,m1),m2) ->
       begin
	 try
	   incr assume_needed_sorts;
	   let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 Prop in
	   decr assume_needed_sorts;
	   let n1l = extract_conjuncts n1 true in
	   List.iter (fun n -> assume_mizar_sorts st g n) n1l;
	   let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 Prop in
	   (Ap(Ap(And,n1),n2),Prop,union_frees frl1 frl2)
	 with
	 | NeedToAssume ->
	     decr assume_needed_sorts;
	     let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 Prop in
	     let n2l = extract_conjuncts n2 true in
	     List.iter (fun n -> assume_mizar_sorts st g n) n2l;
	     let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 Prop in
	     (Ap(Ap(And,n1),n2),Prop,union_frees frl1 frl2)
       end
  | PAp(PAp(PNIff,m1),m2) ->
      let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 Prop in
      let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 Prop in
      (Ap(Ap(Imp,Ap(Ap(Imp,n1),n2)),False),Prop,union_frees frl1 frl2)
  | PNIff ->
      (Lam(Prop,Lam(Prop,preneg (preiff (DB(1,Prop)) (DB(0,Prop))))),ooo,[])
  | PImplies ->
      (Lam(Prop,Lam(Prop,imp (DB(1,Prop)) (DB(0,Prop)))),ooo,[])
  | PRevImplies ->
      (Lam(Prop,Lam(Prop,imp (DB(0,Prop)) (DB(1,Prop)))),ooo,[])
  | PNOr ->
      (Lam(Prop,Lam(Prop,preneg (predisj (DB(1,Prop)) (DB(0,Prop))))),ooo,[])
  | PNAnd ->
      (Lam(Prop,Lam(Prop,preneg (preconj (DB(1,Prop)) (DB(0,Prop))))),ooo,[])
  | PAp(PAp(PEq,m1),m2) -> (*** assume equality is always between type terms of type $i ***)
      let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 istp in
      let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 istp in
      ((eq istp n1 n2),Prop,union_frees frl1 frl2)
  | PEq -> raise (GenericSyntaxError ("Cannot determine type of unapplied equality"));
  | PAp(PAp(PNEq,m1),m2) ->
      let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 istp in
      let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 istp in
      (neg (eq istp n1 n2),Prop,union_frees frl1 frl2)
  | PNEq -> raise (GenericSyntaxError ("Cannot determine type of unapplied negated equality"));
  | PAp(PForall,PLam(xl,m1)) -> let (n,frl) = mizar_elaborate_trm_all st g xl m1 in (n,Prop,frl)
  | PAp(m1,m2) ->
      let (h,ml) = pre_head_spine (PAp(m1,m2)) [] in
      mizar_elaborate_spine st g h ml []
  | PULam(xl,m1) -> raise (GenericSyntaxError("Cannot infer type omitted from lambda"))
  | PLam(xl,m1) -> let (n,ntp,frl) = mizar_elaborate_trm_lam st g xl m1 in (n,ntp,frl)
  | PAll(xl,m1) -> let (n,frl) = mizar_elaborate_trm_all st g xl m1 in (n,Prop,frl)
  | PEx(xl,m1) -> let (n,frl) = mizar_elaborate_trm_exists st g xl m1 in (n,Prop,frl)
  | PExU(x,a1,m1) ->
      begin
	let xaa = to_stp a1 in
	let xx = nxtfreshvar xaa st in
	let (n1,frl1) = mizar_elaborate_trm_force_stp st ((x,xx)::g) m1 Prop in
	(Ap(Lam(Ar(xaa,Prop),Ap(Exists(xaa),Lam(xaa,Ap(Ap(And,Ap(DB(1,Ar(xaa,Prop)),DB(0,xaa))),Ap(Forall(xaa),Lam(xaa,Ap(Ap(Imp,Ap(DB(2,Ar(xaa,Prop)),DB(0,xaa))),Ap(Ap(Eq(xaa),DB(0,xaa)),DB(1,xaa))))))))),Lam(xaa,n1)),Prop,frl1)
      end
  | PChoice((x,xa),m1) ->
      let xaa = to_stp xa in
      let xx = nxtfreshvar xaa st in
      let (n1,frl1) = mizar_elaborate_trm_force_stp st ((x,xx)::g) m1 Prop in
      (Ap(Choice xaa,Lam(xaa,n1)),xaa,frl1)
  | POf(m1,m2) ->
      let b = to_stp m2 in
      let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 b in
      (n1,b,frl1)
  | PDef(m1,_) ->
      mizar_elaborate_trm st g m1
  | PLet(x,m1,m2) ->
      begin
	let (m1a,aa,frl1) = mizar_elaborate_trm st g m1 in
	let (m2b,bb,frl2) = mizar_elaborate_trm st g m2 in
	(Ap(Lam(aa,m2b),m1a),bb,union_frees frl1 frl2)
      end
  | PTLet(x,a1,m1,m2) ->
      begin
	let aa = to_stp a1 in
	let (m1a,frl1) = mizar_elaborate_trm_force_stp st g m1 aa in
	let (m2b,bb,frl2) = mizar_elaborate_trm st g m2 in
	(Ap(Lam(aa,m2b),m1a),bb,union_frees frl1 frl2)
      end
  | _ -> raise (GenericSyntaxError ("Ill-formed term " ^ (pretrm_str m)))
and mizar_elaborate_trm_force_stp (st:mizstate) (g:(string * trm) list) m a : trm * ((string * (trm list)) list) =
  if debug then
    begin
      Printf.printf "mizar_elaborate_trm_force_stp %s : %s\n" (pretrm_str m) (stp_str a);
      Printf.printf "g:\n";
      List.iter (fun (y,n) -> Printf.printf "%s ~> %s\n" y (trm_str n)) g
    end;
  let (ta,_,_,_,_,(_,u,_)) = st in
  try
    let (n,b,frl) = Hashtbl.find ta m in
    if debug then Printf.printf "it's in the table so calling _2\n";
    if not (frl = []) then u := true;
    mizar_elaborate_trm_force_stp_2 n b frl a
  with Not_found ->
    if debug then Printf.printf "it's not in the table so calling _a\n";
    let (n,frl) = mizar_elaborate_trm_force_stp_a st g m a in
    if debug then Printf.printf "back from force_stp_a with n = %s and |frl| = %d\n" (trm_str n) (List.length frl);
    Hashtbl.add ta m (n,a,frl);
    (n,frl)
and mizar_elaborate_trm_force_stp_a (st:mizstate) (g:(string * trm) list) m a : trm * ((string * (trm list)) list) =
  match m with
  | PTrm(n) ->
      if tpof n = a then
	begin
	  let frl = ref [] in
	  let (_,_,_,_,_,(_,u,_)) = st in
	  List.iter
	    (fun (x,a) ->
	      if unknown_name_p st g x then
		begin
		  if not (a = Base("$i")) then raise (ElaborationFailure("Unknown name " ^ x));
		  if !raiseunknownname then raise (UnknownName(x));
		  if debug then
		    begin
		      Printf.printf "#+#++ %s is being treated as a free var (4)\n" x;
		      Printf.printf "g:\n";
		      List.iter (fun (y,n) -> Printf.printf "%s ~> %s\n" y (trm_str n)) g;
		    end;
		  u := true;
		  frl := (x,[])::!frl
		end)
	    (consts_of_trm [] n);
	  (n,!frl)
	end
      else
	raise (ElaborationSimpleTypeError(n,tpof n,a))
  | PAp(PAp(PImplies,m1),m2) ->
      if a = Prop then
	begin
	  let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 Prop in
	  let n1l = extract_conjuncts n1 true in
	  if debug then (Printf.printf "Assuming %d sorting hyps\n" (List.length n1l); List.iter (fun n -> Printf.printf "%s\n" (trm_str n)) n1l);
	  List.iter (fun n -> assume_mizar_sorts st g n) n1l;
	  let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 Prop in
	  (Ap(Ap(Imp,n1),n2),union_frees frl1 frl2)
	end
      else
	raise (ElaborationSimpleTypeError(Name(pretrm_str m,Prop),Prop,a)) (** ridiculous hack to turn m into a trm for the exception **)
  | PAp(PAp(PAnd,m1),m2) ->
      if a = Prop then
	begin
	 try
	   incr assume_needed_sorts;
	   let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 Prop in
	   decr assume_needed_sorts;
	   let n1l = extract_conjuncts n1 true in
	   List.iter (fun n -> assume_mizar_sorts st g n) n1l;
	   let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 Prop in
	   (Ap(Ap(And,n1),n2),union_frees frl1 frl2)
	 with
	 | NeedToAssume ->
	     decr assume_needed_sorts;
	     let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 Prop in
	     let n2l = extract_conjuncts n2 true in
	     List.iter (fun n -> assume_mizar_sorts st g n) n2l;
	     let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 Prop in
	     (Ap(Ap(And,n1),n2),union_frees frl1 frl2)
	end
      else
	raise (ElaborationSimpleTypeError(Name(pretrm_str m,Prop),Prop,a)) (** ridiculous hack to turn m into a trm for the exception **)
  | PAp(PAp(PName "sort",m1),m2) ->
      if a = Prop then
	let (h,ml) = pre_head_spine m2 [] in
	let (n,b,frl) = mizar_elaborate_spine st g h (m1::ml) [] in
	if b = Prop then
	  (n,frl)
	else
	  raise (ElaborationSimpleTypeError(Name(pretrm_str m,Prop),b,Prop)) (** ridiculous hack to turn m into a trm for the exception **)
      else
	raise (ElaborationSimpleTypeError(Name(pretrm_str m,Prop),Prop,a)) (** ridiculous hack to turn m into a trm for the exception **)
  | PAp(PExists,PLam(xl,m1)) ->
      if a = Prop then
	let (n,frl) = mizar_elaborate_trm_exists st g xl m1 in (n,frl)
      else
	raise (ElaborationSimpleTypeError(Name(pretrm_str m,Prop),Prop,a)) (** ridiculous hack to turn m into a trm for the exception **)
  | PAp(PAp(PNIff,m1),m2) ->
      if a = Prop then
	let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 Prop in
	let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 Prop in
	(Ap(Ap(Imp,Ap(Ap(Imp,n1),n2)),False),union_frees frl1 frl2)
      else
	raise (ElaborationSimpleTypeError(Name(pretrm_str m,Prop),Prop,a)) (** ridiculous hack to turn m into a trm for the exception **)
  | PAp(PAp(PEq,m1),m2) -> (*** assume equality is always between type terms of type $i ***)
      if a = Prop then
	let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 istp in
	let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 istp in
	((eq istp n1 n2),union_frees frl1 frl2)
      else
	raise (ElaborationSimpleTypeError(Name(pretrm_str m,Prop),Prop,a)) (** ridiculous hack to turn m into a trm for the exception **)
  | PAp(PAp(PNEq,m1),m2) ->
      if a = Prop then
	let (n1,frl1) = mizar_elaborate_trm_force_stp st g m1 istp in
	let (n2,frl2) = mizar_elaborate_trm_force_stp st g m2 istp in
	(neg (eq istp n1 n2),union_frees frl1 frl2)
      else
	raise (ElaborationSimpleTypeError(Name(pretrm_str m,Prop),Prop,a)) (** ridiculous hack to turn m into a trm for the exception **)
  | PAp(PForall,PLam(xl,m1)) ->
      if a = Prop then
	let (n,frl) = mizar_elaborate_trm_all st g xl m1 in (n,frl)
      else
	raise (ElaborationSimpleTypeError(Name(pretrm_str m,Prop),Prop,a)) (** ridiculous hack to turn m into a trm for the exception **)
  | PAp(m1,m2) ->
      if a = Base("$i") then
        begin
          let (h,ml) = pre_head_spine (PAp(m1,m2)) [] in
	  let (n,b,frl) = mizar_elaborate_spine st g h ml [] in
	  mizar_elaborate_trm_force_stp_2 n b frl a
        end
      else if a = Prop then
        begin
          let (h,ml) = pre_head_spine (PAp(m1,m2)) [] in
	  let (n,b,frl) = mizar_elaborate_spine st g h ml [] in
	  mizar_elaborate_trm_force_stp_2 n b frl a
        end
      else
        begin
          let (h,ml) = pre_head_spine (PAp(m1,m2)) [] in
          match h with
          | PName x ->
	      begin
		try
		  if debug then (Printf.printf "x : %s\n" x; flush stdout);
                  let xa = Hashtbl.find !mizar_name_tp x in
		  if debug then (Printf.printf "xa : %s\n" (stp_str xa); flush stdout);
                  let (n,b,frl) = mizar_elaborate_spine_other st g (Name(x,xa)) xa ml [] in
                  mizar_elaborate_trm_force_stp_2 n b frl a
		with Not_found -> raise (ElaborationFailure("Unknown applied name " ^ x))
	      end
          | _ ->
	      let (mh,mhtp,frl) = mizar_elaborate_trm st g h in
	      let (n,b,frl) = mizar_elaborate_spine_other st g mh mhtp ml frl in
	      mizar_elaborate_trm_force_stp_2 n b frl a
        end
  | _ ->
      begin
	let (n,b,frl) = mizar_elaborate_trm st g m in
	mizar_elaborate_trm_force_stp_2 n b frl a
      end
and mizar_elaborate_trm_force_stp_2 m b frl a : trm * ((string * (trm list)) list) =
  if a = b then
    (m,frl)
  else
    raise (ElaborationSimpleTypeError(m,b,a))
and mizar_elaborate_trm_lam (st:mizstate) (g:(string * trm) list) xl m : trm * stp * ((string * (trm list)) list) =
  match xl with
    ((x,xa)::xr) ->
      begin
	let xaa = to_stp xa in
	let xx = nxtfreshvar xaa st in
	let (n1,b,frl1) = mizar_elaborate_trm_lam st ((x,xx)::g) xr m in
	(Lam(xaa,abstract_subterms 0 xx n1),Ar(xaa,b),frl1)
      end
  | [] ->
      mizar_elaborate_trm st g m
and mizar_elaborate_trm_all st g xl m : trm * ((string * (trm list)) list) =
  match xl with
    ((x,xa)::xr) ->
      begin
	let xaa = to_stp xa in
	let xx = nxtfreshvar xaa st in
	let (n1,frl1) = mizar_elaborate_trm_all st ((x,xx)::g) xr m in
	(Ap(Forall(xaa),Lam(xaa,abstract_subterms 0 xx n1)),frl1)
      end
  | [] ->
      mizar_elaborate_trm_force_stp st g m Prop
and mizar_elaborate_trm_exists st g xl m : trm * ((string * (trm list)) list) =
  match xl with
    ((x,xa)::xr) ->
      begin
	let xaa = to_stp xa in
	let xx = nxtfreshvar xaa st in
	let (n1,frl1) = mizar_elaborate_trm_exists st ((x,xx)::g) xr m in
	(Ap(Exists(xaa),Lam(xaa,abstract_subterms 0 xx n1)),frl1)
      end
  | [] ->
      mizar_elaborate_trm_force_stp st g m Prop
and mizar_elaborate_trm_var x g : trm * stp =
  match g with
    ((y,yy)::gr) -> 
      if (x = y) then
	(yy,tpof yy)
      else
	mizar_elaborate_trm_var x gr
  | [] -> raise Not_found


